import asyncio
import time
from typing import Any, TypeVar

from frappy.client import CacheItem, Logger, SecopClient

T = TypeVar("T")


class AsyncFrappyClient:
    def __init__(self, host: str, port: str, loop) -> None:
        self.host: str = host
        self.port: str = port

        self.client: SecopClient = None

        self.loop = loop

        self.external: bool = False

        self.conn_timestamp: float

        self.log = None

    @property
    def state(self):
        return self.client.state

    @property
    def online(self):
        return self.client.online

    @property
    def modules(self):
        return self.client.modules

    @property
    def properties(self):
        return self.client.properties

    @property
    def uri(self):
        return self.client.uri

    @property
    def nodename(self):
        return self.client.nodename

    @classmethod
    async def create(cls, host, port, loop, log=Logger):
        self = AsyncFrappyClient(host=host, port=port, loop=loop)
        self.client = SecopClient(uri=host + ":" + port, log=log)

        self.log = self.client.log

        await self.connect(3)

        return self

    async def connect(self, try_period=0):
        await asyncio.to_thread(self.client.connect, try_period)
        self.conn_timestamp = time.time()

    async def disconnect(self, shutdown=True):
        await asyncio.to_thread(self.client.disconnect, shutdown)

    async def get_parameter(self, module, parameter, trycache=False) -> CacheItem:
        paramerter_reading = await asyncio.to_thread(
            self.client.getParameter, module, parameter, trycache
        )
        return paramerter_reading

    async def set_parameter(self, module, parameter, value) -> CacheItem:
        paramerter_reading = await asyncio.to_thread(
            self.client.setParameter, module, parameter, value
        )
        return paramerter_reading

    async def exec_command(self, module, command, argument=None) -> tuple[Any, dict]:
        return await asyncio.to_thread(
            self.client.execCommand, module, command, argument
        )

    def register_callback(self, key, *args, **kwds):
        self.client.register_callback(key, *args, **kwds)

    def unregister_callback(self, key, *args, **kwds):
        self.client.unregister_callback(key, *args, **kwds)



# mypy: disable-error-code="attr-defined"
import os

from secop_ophyd.SECoPDevices import SECoPNodeDevice


async def test_class_gen(nested_struct_sim, nested_node: SECoPNodeDevice):
    nested_node.class_from_instance()

    if os.path.exists("genNodeClass.py"):
        os.remove("genNodeClass.py")

    await nested_node.disconnect_async()


async def test_class_gen_path(nested_struct_sim, nested_node: SECoPNodeDevice):
    nested_node.class_from_instance("tests")

    if os.path.exists("tests/genNodeClass.py"):
        os.remove("tests/genNodeClass.py")

    await nested_node.disconnect_async()



#  -*- coding: utf-8 -*-
# *****************************************************************************
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Enrico Faulhaber <enrico.faulhaber@frm2.tum.de>
#
# *****************************************************************************
"""playing implementation of a (simple) simulated cryostat"""


import random
import time
from math import atan

from frappy.datatypes import BoolType, EnumType, FloatRange, StringType, TupleOf
from frappy.lib import clamp, mkthread
from frappy.modules import Command, Drivable, Parameter

# test custom property (value.test can be changed in config file)
from frappy.properties import Property


class TestParameter(Parameter):
    test = Property(
        "A Property for testing purposes", StringType(), default="", export=True
    )


class CryoBase(Drivable):
    is_cryo = Property(
        "private Flag if this is a cryostat", BoolType(), default=True, export=True
    )


class Cryostat(CryoBase):
    """simulated cryostat with:

    - heat capacity of the sample
    - cooling power
    - thermal transfer between regulation and samplen
    """

    jitter = Parameter(
        "amount of random noise on readout values",
        datatype=FloatRange(0, 1),
        unit="K",
        default=0.1,
        readonly=False,
        export=False,
    )
    T_start = Parameter(
        "starting temperature for simulation",
        datatype=FloatRange(0),
        default=10,
        export=False,
    )
    looptime = Parameter(
        "timestep for simulation",
        datatype=FloatRange(0.01, 10),
        unit="s",
        default=1,
        readonly=False,
        export=False,
    )
    ramp = Parameter(
        "ramping speed of the setpoint",
        datatype=FloatRange(0, 1e3),
        unit="K/min",
        default=1,
        readonly=False,
    )
    setpoint = Parameter(
        "current setpoint during ramping else target",
        datatype=FloatRange(),
        default=1,
        unit="K",
    )
    maxpower = Parameter(
        "Maximum heater power",
        datatype=FloatRange(0),
        default=1,
        unit="W",
        readonly=False,
        group="heater_settings",
    )
    heater = Parameter(
        "current heater setting",
        datatype=FloatRange(0, 100),
        default=0,
        unit="%",
        group="heater_settings",
    )
    heaterpower = Parameter(
        "current heater power",
        datatype=FloatRange(0),
        default=0,
        unit="W",
        group="heater_settings",
    )
    target = Parameter(
        "target temperature",
        datatype=FloatRange(0),
        default=0,
        unit="K",
        readonly=False,
    )
    value = TestParameter(
        "regulation temperature",
        datatype=FloatRange(0),
        default=0,
        unit="K",
        test="TEST",
    )
    pid = Parameter(
        "regulation coefficients",
        datatype=TupleOf(FloatRange(0), FloatRange(0, 100), FloatRange(0, 100)),
        default=(40, 10, 2),
        readonly=False,
        group="pid",
    )
    # pylint: disable=invalid-name
    p = Parameter(
        "regulation coefficient 'p'",
        datatype=FloatRange(0),
        default=40,
        unit="%/K",
        readonly=False,
        group="pid",
    )
    i = Parameter(
        "regulation coefficient 'i'",
        datatype=FloatRange(0, 100),
        default=10,
        readonly=False,
        group="pid",
    )
    d = Parameter(
        "regulation coefficient 'd'",
        datatype=FloatRange(0, 100),
        default=2,
        readonly=False,
        group="pid",
    )
    mode = Parameter(
        "mode of regulation",
        datatype=EnumType("mode", ramp=None, pid=None, openloop=None),
        default="ramp",
        readonly=False,
    )
    pollinterval = Parameter("polling interval", datatype=FloatRange(0), default=5)
    tolerance = Parameter(
        "temperature range for stability checking",
        datatype=FloatRange(0, 100),
        default=0.1,
        unit="K",
        readonly=False,
        group="stability",
    )
    window = Parameter(
        "time window for stability checking",
        datatype=FloatRange(1, 900),
        default=30,
        unit="s",
        readonly=False,
        group="stability",
    )
    timeout = Parameter(
        "max waiting time for stabilisation check",
        datatype=FloatRange(1, 36000),
        default=900,
        unit="s",
        readonly=False,
        group="stability",
    )

    def initModule(self):
        super().initModule()
        self._stopflag = False
        self._thread = mkthread(self.thread)

    def read_status(self):
        # instead of asking a 'Hardware' take the value from the simulation
        return self.status

    def read_value(self):
        # return regulation value (averaged regulation temp)
        return self.regulationtemp + self.jitter * (0.5 - random.random())

    def read_target(self):
        return self.target

    def write_target(self, value):
        value = round(value, 2)
        if value == self.target:
            # nothing to do
            return value
        self.target = value
        # next read_status will see this status, until the loop updates it
        self.status = self.Status.BUSY, "new target set"
        return value

    def read_maxpower(self):
        return self.maxpower

    def write_maxpower(self, newpower):
        # rescale heater setting in % to keep the power
        heat = max(0, min(100, self.heater * self.maxpower / float(newpower)))
        self.heater = heat
        self.maxpower = newpower
        return newpower

    def write_pid(self, newpid):
        self.p, self.i, self.d = newpid
        return (self.p, self.i, self.d)

    def read_pid(self):
        return (self.p, self.i, self.d)

    @Command()
    def stop(self):
        """Stop ramping the setpoint

        by setting the current setpoint as new target"""
        # XXX: discussion: take setpoint or current value ???
        self.write_target(self.setpoint)

    #
    # calculation helpers
    #
    def __coolerPower(self, temp):
        """returns cooling power in W at given temperature"""
        # quadratic up to 42K, is linear from 40W@42K to 100W@600K
        # return clamp((temp-2)**2 / 32., 0., 40.) + temp * 0.1
        return clamp(15 * atan(temp * 0.01) ** 3, 0.0, 40.0) + temp * 0.1 - 0.2

    def __coolerCP(self, temp):
        """heat capacity of cooler at given temp"""
        return 75 * atan(temp / 50) ** 2 + 1

    def __heatLink(self, coolertemp, sampletemp):
        """heatflow from sample to cooler. may be negative..."""
        flow = (sampletemp - coolertemp) * ((coolertemp + sampletemp) ** 2) / 400.0
        cp = clamp(self.__coolerCP(coolertemp) * self.__sampleCP(sampletemp), 1, 10)
        return clamp(flow, -cp, cp)

    def __sampleCP(self, temp):
        return 3 * atan(temp / 30) + 12 * temp / ((temp - 12.0) ** 2 + 10) + 0.5

    def __sampleLeak(self, temp):
        return 0.02 / temp

    def thread(self):
        self.sampletemp = self.T_start
        self.regulationtemp = self.T_start
        self.status = self.Status.IDLE, ""
        while not self._stopflag:
            try:
                self.__sim()
            except Exception as e:
                self.log.exception(e)
                self.status = self.Status.ERROR, str(e)

    def __sim(self):
        # complex thread handling:
        # a) simulation of cryo (heat flow, thermal masses,....)
        # b) optional PID temperature controller with windup control
        # c) generating status+updated value+ramp
        # this thread is not supposed to exit!

        self.setpoint = self.target
        # local state keeping:
        regulation = self.regulationtemp
        sample = self.sampletemp
        # keep history values for stability check
        window = []
        timestamp = time.time()
        heater = 0
        lastflow = 0
        last_heaters = (0, 0)
        delta = 0
        _I = _D = 0
        lastD = 0
        damper = 1
        lastmode = self.mode
        while not self._stopflag:
            t = time.time()
            h = t - timestamp
            if h < self.looptime / damper:
                time.sleep(clamp(self.looptime / damper - h, 0.1, 60))
                continue
            # a)
            sample = self.sampletemp
            regulation = self.regulationtemp
            heater = self.heater

            heatflow = self.__heatLink(regulation, sample)
            self.log.debug(
                "sample = %.5f, regulation = %.5f, heatflow = %.5g",
                sample,
                regulation,
                heatflow,
            )
            newsample = max(
                0,
                sample
                + (self.__sampleLeak(sample) - heatflow) / self.__sampleCP(sample) * h,
            )
            # avoid instabilities due to too small CP
            newsample = clamp(newsample, sample, regulation)
            regdelta = (
                heater * 0.01 * self.maxpower
                + heatflow
                - self.__coolerPower(regulation)
            )
            newregulation = max(
                0, regulation + regdelta / self.__coolerCP(regulation) * h
            )
            # b) see
            # http://brettbeauregard.com/blog/2011/04/
            # improving-the-beginners-pid-introduction/
            if self.mode != self.mode.openloop:
                # fix artefacts due to too big timesteps
                # actually i would prefer reducing looptime, but i have no
                # good idea on when to increase it back again
                if heatflow * lastflow != -100:
                    if (newregulation - newsample) * (regulation - sample) < 0:
                        # newregulation = (newregulation + regulation) / 2
                        # newsample = (newsample + sample) / 2
                        damper += 1
                lastflow = heatflow

                error = self.setpoint - newregulation
                # use a simple filter to smooth delta a little
                delta = (delta + regulation - newregulation) * 0.5

                kp = self.p * 0.1  # LakeShore P = 10*k_p
                ki = kp * abs(self.i) / 500.0  # LakeShore I = 500/T_i
                kd = kp * abs(self.d) / 2.0  # LakeShore D = 2*T_d
                _P = kp * error
                _I += ki * error * h
                _D = kd * delta / h

                # avoid reset windup
                _I = clamp(_I, 0.0, 100.0)  # _I is in %

                # avoid jumping heaterpower if switching back to pid mode
                if lastmode != self.mode:
                    # adjust some values upon switching back on
                    _I = self.heater - _P - _D

                v = _P + _I + _D
                # in damping mode, use a weighted sum of old + new heaterpower
                if damper > 1:
                    v = ((damper**2 - 1) * self.heater + v) / damper**2

                # damp oscillations due to D switching signs
                if _D * lastD < -0.2:
                    v = (v + heater) * 0.5
                # clamp new heater power to 0..100%
                heater = clamp(v, 0.0, 100.0)
                lastD = _D

                self.log.debug(
                    "PID: P = %.2f, I = %.2f, D = %.2f, " "heater = %.2f",
                    _P,
                    _I,
                    _D,
                    heater,
                )

                # check for turn-around points to detect oscillations ->
                # increase damper
                x, y = last_heaters
                if (x + 0.1 < y and y > heater + 0.1) or (
                    x > y + 0.1 and y + 0.1 < heater
                ):
                    damper += 1
                last_heaters = (y, heater)

            else:
                # self.heaterpower is set manually, not by pid
                heater = self.heater
                last_heaters = (0, 0)

            heater = round(heater, 1)
            sample = newsample
            regulation = newregulation
            lastmode = self.mode
            # c)
            if self.setpoint != self.target:
                if self.ramp == 0 or self.mode == self.mode.enum.pid:
                    maxdelta = 10000
                else:
                    maxdelta = self.ramp / 60.0 * h
                try:
                    self.setpoint = round(
                        self.setpoint
                        + clamp(self.target - self.setpoint, -maxdelta, maxdelta),
                        3,
                    )
                    self.log.debug(
                        "setpoint changes to %r (target %r)", self.setpoint, self.target
                    )
                except (TypeError, ValueError):
                    # self.target might be None
                    pass

            # temperature is stable when all recorded values in the window
            # differ from setpoint by less than tolerance
            currenttime = time.time()
            window.append((currenttime, sample))
            while window[0][0] < currenttime - self.window:
                # remove old/stale entries
                window.pop(0)
            # obtain min/max
            deviation = 0
            for _, _T in window:
                if abs(_T - self.target) > deviation:
                    deviation = abs(_T - self.target)
            if (len(window) < 3) or deviation > self.tolerance:
                self.status = self.Status.BUSY, "unstable"
            elif self.setpoint == self.target:
                self.status = self.Status.IDLE, "at target"
                damper -= (damper - 1) * 0.1  # max value for damper is 11
            else:
                self.status = self.Status.BUSY, "ramping setpoint"
            damper -= (damper - 1) * 0.05
            self.regulationtemp = round(regulation, 3)
            self.sampletemp = round(sample, 3)
            self.heaterpower = round(heater * self.maxpower * 0.01, 3)
            self.heater = heater
            timestamp = t
            self.read_value()

    def shutdownModule(self):
        # should be called from server when the server is stopped
        self._stopflag = True
        if self._thread and self._thread.is_alive():
            self._thread.join()



# mypy: disable-error-code="attr-defined"
import asyncio

from frappy.client import CacheItem

# import xprocess
from secop_ophyd.AsyncFrappyClient import AsyncFrappyClient


async def test_asycnc_secopclient_conn(
    cryo_sim, async_frappy_client: AsyncFrappyClient
):
    assert async_frappy_client.online is True
    await async_frappy_client.disconnect()


async def test_asycnc_secopclient_get_param(
    cryo_sim, async_frappy_client: AsyncFrappyClient
):
    reading = await async_frappy_client.get_parameter("cryo", "value", False)

    assert isinstance(reading, CacheItem)

    await async_frappy_client.disconnect()


async def test_async_secopclient_disconnect(
    cryo_sim, async_frappy_client: AsyncFrappyClient
):
    await async_frappy_client.get_parameter("cryo", "value", False)

    await async_frappy_client.disconnect(True)

    assert async_frappy_client.state == "shutdown"


async def test_async_secopclient_reconn(
    cryo_sim, async_frappy_client: AsyncFrappyClient
):
    reading1: CacheItem = await async_frappy_client.get_parameter(
        "cryo", "value", False
    )
    reading2: CacheItem = await async_frappy_client.get_parameter(
        "cryo", "value", False
    )

    assert reading1.value != reading2.value

    await async_frappy_client.disconnect(False)

    # for a short period the status is still "connected"
    # (the disconn task finishes and the state is only set to a new value
    # once the reconnect thread starts)
    while async_frappy_client.state == "connected":
        await asyncio.sleep(0.001)

    while async_frappy_client.state == "reconnecting":
        await asyncio.sleep(0.001)

    while async_frappy_client.state == "activating":
        await asyncio.sleep(0.001)

    assert async_frappy_client.state == "connected"

    # ensures we are connected and getting fresh data again
    reading3: CacheItem = await async_frappy_client.get_parameter(
        "cryo", "value", False
    )
    reading4: CacheItem = await async_frappy_client.get_parameter(
        "cryo", "value", False
    )

    assert reading3.value != reading4.value

    await async_frappy_client.disconnect(True)


async def test_async_secopclient_shutdown_and_reconn(
    cryo_sim, async_frappy_client: AsyncFrappyClient
):
    reading1: CacheItem = await async_frappy_client.get_parameter(
        "cryo", "value", False
    )
    reading2: CacheItem = await async_frappy_client.get_parameter(
        "cryo", "value", False
    )

    assert reading1.value != reading2.value

    # Shutdown
    await async_frappy_client.disconnect(True)

    assert async_frappy_client.state == "shutdown"

    await asyncio.sleep(2)

    # ensure no auto reconnect
    assert async_frappy_client.state == "shutdown"

    # manual reconn
    async_frappy_client.client._shutdown.clear()
    await async_frappy_client.connect(3)

    assert async_frappy_client.state == "connected"

    # ensures we are connected and getting fresh data again
    reading1 = await async_frappy_client.get_parameter("cryo", "value", False)
    reading2 = await async_frappy_client.get_parameter("cryo", "value", False)

    assert reading1.value != reading2.value

    await async_frappy_client.disconnect(True)



# mypy: disable-error-code="attr-defined"
import asyncio
import logging
import os

import pytest
from bluesky import RunEngine
from databroker.v2 import temp
from frappy.datatypes import (
    ArrayOf,
    FloatRange,
    IntRange,
    StringType,
    StructOf,
    TupleOf,
)
from xprocess import ProcessStarter

from secop_ophyd.AsyncFrappyClient import AsyncFrappyClient
from secop_ophyd.SECoPDevices import SECoPNodeDevice


@pytest.fixture
def frappy_env():
    """Create and return environment variables and paths for frappy server."""
    # Create absolute paths for frappy requirements
    base_dir = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))
    conf_dir = os.path.join(base_dir, "cfg")
    log_dir = os.path.join(base_dir, "logs")
    pid_dir = os.path.join(base_dir, "pid")

    # Add path to frappy_modules
    frappy_modules_dir = os.path.join(base_dir, "frappy_modules")

    # Create environment variables
    env = os.environ.copy()  # Start with current environment

    env.update(
        {
            "FRAPPY_CONFDIR": conf_dir,
            "FRAPPY_LOGDIR": log_dir,
            "FRAPPY_PIDDIR": pid_dir,
            "FRAPPY_PROJECT_ROOT": base_dir,  # Add the project root
        }
    )

    return {
        "env": env,
        "conf_dir": conf_dir,
        "log_dir": log_dir,
        "pid_dir": pid_dir,
        "modules_dir": frappy_modules_dir,
    }


@pytest.fixture
def cryo_sim(xprocess, frappy_env):

    class Starter(ProcessStarter):
        # startup pattern
        pattern = ".*: startup done with interface(s)*"
        timeout = 5
        # command to start process
        env = frappy_env["env"]  # Use the environment variables from the fixture

        args = [
            "frappy-server",
            "-c",
            "cryo_cfg.py",
            "cryo",
        ]

    # ensure process is running and return its logfile
    xprocess.ensure("cryo_sim", Starter)

    yield xprocess

    # clean up whole process tree afterwards
    xprocess.getinfo("cryo_sim").terminate()


@pytest.fixture
def nested_struct_sim(xprocess, frappy_env):

    class Starter(ProcessStarter):
        # startup pattern
        pattern = ".*: startup done with interface(s)*"
        timeout = 5

        env = frappy_env["env"]  # Use the environment variables from the fixture
        args = [
            "frappy-server",
            "-c",
            "ophyd_secop_test_cfg.py",
            "nested",
        ]

    pname = "nested"
    # ensure process is running and return its logfile
    xprocess.ensure(pname, Starter)

    yield

    # clean up whole process tree afterwards
    xprocess.getinfo(pname).terminate()


@pytest.fixture
def logger():
    class NoRXFilter(logging.Filter):
        def filter(self, record):
            return not record.getMessage().startswith("RX")

    class NoTXFilter(logging.Filter):
        def filter(self, record):
            return not record.getMessage().startswith("TX")

    logger = logging.getLogger("async_client")
    logger.setLevel(logging.DEBUG)

    logger.addFilter(NoRXFilter())
    logger.addFilter(NoTXFilter())

    logging.basicConfig(
        datefmt="%d/%m/%Y %I:%M:%S",
        format="%(asctime)s.%(msecs)03d  %(name)-12s %(levelname)-8s %(message)s",
        filename="asyncclient.log",
        encoding="utf-8",
        level=logging.DEBUG,
        filemode="w",
    )

    return logger


@pytest.fixture
async def async_frappy_client(cryo_sim, logger, port="10769"):
    loop = asyncio.get_running_loop()

    return await AsyncFrappyClient.create(
        host="localhost", port=port, loop=loop, log=logger
    )


@pytest.fixture
async def nested_client(nested_struct_sim, logger, port="10771"):
    loop = asyncio.get_running_loop()

    return await AsyncFrappyClient.create(
        host="localhost", port=port, loop=loop, log=logger
    )


@pytest.fixture
def db():
    return temp()


@pytest.fixture
def run_engine(db):
    re = RunEngine({})
    re.subscribe(db.v1.insert)
    return re


@pytest.fixture
def cryo_node(run_engine):
    return SECoPNodeDevice.create(host="localhost", port="10769", loop=run_engine.loop)


@pytest.fixture
def nested_node_re(run_engine):
    return SECoPNodeDevice.create(host="localhost", port="10771", loop=run_engine.loop)


@pytest.fixture
async def cryo_node_internal_loop():
    return await SECoPNodeDevice.create_async(
        host="localhost", port="10769", loop=asyncio.get_running_loop()
    )


@pytest.fixture
async def nested_node():
    return await SECoPNodeDevice.create_async(
        host="localhost", port="10771", loop=asyncio.get_running_loop()
    )


@pytest.fixture
def struct_val():
    return {
        "number": 5.0,
        "string": "blablalbla",
        "tupl": [1.0, 1.0, 1.0],
        "pos_struct": {"x": 5.0, "y": 10.0, "z": 15.0, "col": "green", "enum": 1},
    }


@pytest.fixture
def nested_param_description():
    return {
        "_nested_struct": {
            "datainfo": {
                "members": {
                    "number": {"max": 100.0, "min": 0.0, "type": "double", "unit": "s"},
                    "pos_struct": {
                        "members": {
                            "col": {"type": "string"},
                            "enum": {
                                "members": {
                                    "mode_max": 2,
                                    "mode_one": 1,
                                    "mode_zero": 0,
                                },
                                "type": "enum",
                            },
                            "x": {
                                "max": 100.0,
                                "min": 0.0,
                                "type": "double",
                                "unit": "m",
                            },
                            "y": {
                                "max": 100.0,
                                "min": 0.0,
                                "type": "double",
                                "unit": "m",
                            },
                            "z": {
                                "max": 100.0,
                                "min": 0.0,
                                "type": "double",
                                "unit": "m",
                            },
                        },
                        "type": "struct",
                    },
                    "string": {"type": "string"},
                    "tupl": {
                        "members": [
                            {"max": 100.0, "min": 0.0, "type": "double"},
                            {"max": 100.0, "min": 0.0, "type": "double"},
                            {"max": 100.0, "min": 0.0, "type": "double"},
                        ],
                        "type": "tuple",
                    },
                },
                "type": "struct",
            },
            "description": "nestedstruct dict containing other structs and tuples ",
            "readonly": True,
        }
    }


@pytest.fixture
def array_dtype():
    dty = StructOf(
        flt=FloatRange(), testarr=ArrayOf(members=IntRange(), minlen=5, maxlen=5)
    )

    elem = {"flt": 1.3, "testarr": [1, 2, 3, 4, 5]}
    dty.check_type(elem)

    return dty, elem


@pytest.fixture
def array_tuple_dtype():
    dty = TupleOf(FloatRange(), ArrayOf(members=IntRange(), minlen=5, maxlen=5))

    elem = {"flt": 1.3, "testarr": [1, 2, 3, 4, 5]}
    dty.check_type(elem)

    return dty, elem


@pytest.fixture
def array_ragged_dtype():
    dty = StructOf(
        flt=FloatRange(),
        testarr=ArrayOf(
            members=ArrayOf(members=IntRange(), minlen=1, maxlen=5), minlen=5, maxlen=5
        ),
    )

    elem = {"flt": 1.3, "testarr": [[1, 2, 3], [1, 2, 3, 4, 5], [1, 2], [1], [1, 2, 5]]}
    dty.check_type(elem)

    return dty, elem


@pytest.fixture
def array_2d_dtype():
    dty = StructOf(
        flt=FloatRange(),
        testarr=ArrayOf(
            members=ArrayOf(members=IntRange(), minlen=5, maxlen=5), minlen=5, maxlen=5
        ),
    )

    elem = {
        "flt": 1.3,
        "testarr": [
            [1, 2, 3, 4, 5],
            [1, 2, 3, 4, 5],
            [1, 2, 3, 4, 5],
            [1, 2, 3, 4, 5],
            [1, 2, 3, 4, 5],
        ],
    }
    dty.check_type(elem)

    return dty, elem


@pytest.fixture
def array_of_structs_dtype():
    dty = StructOf(
        flt=FloatRange(),
        testarr=ArrayOf(
            members=StructOf(
                red=FloatRange(),
                green=FloatRange(),
                blue=FloatRange(),
                name=StringType(),
            ),
            minlen=5,
            maxlen=5,
        ),
    )

    elem = {
        "flt": 1.3,
        "testarr": [
            {"red": 1.2, "green": 3.4, "blue": 24.5, "name": "dieter"},
            {"red": 1.2, "green": 3.4, "blue": 24.5, "name": "dieter"},
            {"red": 1.2, "green": 3.4, "blue": 24.5, "name": "dieter"},
            {"red": 1.2, "green": 3.4, "blue": 24.5, "name": "dieter"},
            {"red": 1.2, "green": 3.4, "blue": 24.5, "name": "dieter"},
        ],
    }
    dty.check_type(elem)

    return dty, elem


@pytest.fixture
def bare_array_of_structs_dtype():
    dty = ArrayOf(
        members=StructOf(
            red=FloatRange(), green=FloatRange(), blue=FloatRange(), name=StringType()
        ),
        minlen=5,
        maxlen=5,
    )

    elem = [
        {"red": 1.2, "green": 3.4, "blue": 24.5, "name": "dieter"},
        {"red": 1.2, "green": 3.4, "blue": 24.5, "name": "dieter"},
        {"red": 1.2, "green": 3.4, "blue": 24.5, "name": "dieter"},
        {"red": 1.2, "green": 3.4, "blue": 24.5, "name": "dieter"},
        {"red": 1.2, "green": 3.4, "blue": 24.5, "name": "dieter"},
    ]
    dty.check_type(elem)

    return dty, elem



import asyncio
import warnings
from functools import wraps
from typing import Any, Callable, Dict, Optional

from bluesky.protocols import DataKey, Reading
from frappy.client import CacheItem
from frappy.datatypes import (
    ArrayOf,
    BLOBType,
    BoolType,
    DataType,
    FloatRange,
    IntRange,
    ScaledInteger,
    StringType,
    StructOf,
    TupleOf,
)
from ophyd_async.core import Callback, SignalBackend, T

from secop_ophyd.AsyncFrappyClient import AsyncFrappyClient
from secop_ophyd.util import Path, SECoPdtype, SECoPReading, deep_get

atomic_dtypes = (
    StringType,
    ScaledInteger,
    IntRange,
    FloatRange,
    BoolType,
    BLOBType,
    ArrayOf,
)


# max depth for datatypes supported by tiled/databroker
MAX_DEPTH = 1


class LocalBackend(SignalBackend):
    """Class for the 'argument' and 'result' Signal backends of a SECoP_CMD_Device.
    These Signals act as a local cache for storing the command argument and result.

    """

    def __init__(
        self, path: Path, secop_dtype_obj: DataType, sig_datainfo: dict
    ) -> None:
        """Initialize SECoP_CMD_IO_Backend

        :param path: Path to the command in the secclient module dict
        :type path: Path
        :param SECoPdtype_obj: detailed SECoP datatype object for bidirectional
        conversion between JSON to and numpy arrays
        :type SECoPdtype_obj: DataType
        :param sig_datainfo: SECoP datainfo string of the value represented
        by the signal
        :type sig_datainfo: dict
        """
        self.SECoP_type_info: SECoPdtype = SECoPdtype(secop_dtype_obj)

        self.reading: SECoPReading = SECoPReading(
            secop_dt=self.SECoP_type_info, entry=None
        )

        # module:acessible Path for reading/writing (module,accessible)
        self.path: Path = path

        # Root datainfo or memberinfo for nested datatypes
        # TODO check if this is really needed
        self.datainfo: dict = sig_datainfo

        self.callback: Callback | None = None

        self.SECoPdtype_obj: DataType = secop_dtype_obj

        self.describe_dict: dict

        self.source_name = self.path._module_name + ":" + self.path._accessible_name

        self.describe_dict = {}

        self.describe_dict["source"] = self.source("", True)

        self.describe_dict.update(self.SECoP_type_info.get_datakey())

        for property_name, prop_val in self.datainfo.items():
            if property_name == "type":
                property_name = "SECoP_dtype"
            self.describe_dict[property_name] = prop_val

        super().__init__(datatype=self.SECoP_type_info.np_datatype)

    def source(self, name: str, read: bool) -> str:
        return self.source_name

    async def connect(self, timeout: float):
        pass

    async def put(self, value: Any | None, wait=True):
        self.reading.set_reading(self.SECoP_type_info.val2secop(value))

        if self.callback is not None:
            self.callback(self.reading.get_reading())

    async def get_datakey(self, source: str) -> DataKey:
        """Metadata like source, dtype, shape, precision, units"""
        return self.describe_dict

    async def get_reading(self) -> Reading:
        return self.reading.get_reading()

    async def get_value(self) -> T:
        return self.reading.get_value()

    async def get_setpoint(self) -> T:
        return await self.get_value()

    def set_callback(self, callback: Callback[T] | None) -> None:
        self.callback = callback  # type: ignore[assignment]


# TODO add return of Asyncstatus
class SECoPXBackend(SignalBackend):
    """
    Signal backend for SignalX of a SECoP_CMD_Device, that handles command execution

    """

    def __init__(
        self,
        path: Path,
        secclient: AsyncFrappyClient,
        argument: LocalBackend | None,
        result: LocalBackend | None,
    ) -> None:
        """Initializes SECoP_CMD_X_Backend

        :param path: Path to the command in the secclient module dict
        :type path: Path
        :param secclient: SECoP client providing communication to the SEC Node
        :type secclient: AsyncFrappyClient
        :param argument: Refence to Argument Signal
        :type argument: SECoP_CMD_IO_Backend | None
        :param result: Reference to Result Signal
        :type result: SECoP_CMD_IO_Backend | None
        """

        self._secclient: AsyncFrappyClient = secclient

        # module:acessible Path for reading/writing (module,accessible)
        self.path: Path = path

        self.callback: Callable
        self.argument: LocalBackend | None = argument
        self.result: LocalBackend | None = result

        self.source_name = self.path._module_name + ":" + self.path._accessible_name
        super().__init__(datatype=None)

    def source(self, name: str, read: bool) -> str:
        return self.source_name

    async def connect(self, timeout: float):
        pass

    async def put(self, value: Any | None, wait=True):

        if self.argument is None:
            argument = None
        else:
            argument = await self.argument.get_value()

        res, qualifiers = await asyncio.wait_for(
            fut=self._secclient.exec_command(
                module=self.path._module_name,
                command=self.path._accessible_name,
                argument=argument,
            ),
            timeout=None,
        )

        # write return Value to corresponding Backend

        if self.result is None:
            return
        else:
            val = self.result.SECoP_type_info.secop2val(res)

            await self.result.put(val)

    async def get_datakey(self, source: str) -> DataKey:
        """Metadata like source, dtype, shape, precision, units"""
        res = {}

        res["source"] = self.source("", True)

        # ophyd datatype (some SECoP datatypeshaveto be converted)
        # signalx has no datatype and is never read
        res["dtype"] = "None"

        # get shape from datainfo and SECoPtype

        res["shape"] = []  # type: ignore

        return res

    async def get_reading(self) -> Reading:
        raise Exception(
            "Cannot read _x Signal, it has no value and is only"
            + " used to trigger Command execution"
        )

    async def get_value(self) -> T:
        raise Exception(
            "Cannot read _x Signal, it has no value and is only"
            + " used to trigger Command execution"
        )

    def set_callback(self, callback: Callback[T] | None) -> None:
        pass

    async def get_setpoint(self) -> T:
        raise Exception(
            "Cannot read _x Signal, it has no value and is only"
            + " used to trigger Command execution"
        )


class SECoPParamBackend(SignalBackend):
    """Standard backend for a Signal that represents SECoP Parameter"""

    def __init__(self, path: Path, secclient: AsyncFrappyClient) -> None:
        """_summary_

        :param path: Path to the parameter in the secclient module dict
        :type path: Path
        :param secclient: SECoP client providing communication to the SEC Node
        :type secclient: AsyncFrappyClient
        """

        # secclient
        self._secclient: AsyncFrappyClient = secclient

        # module:acessible Path for reading/writing (module,accessible)
        self.path: Path = path

        self._param_description: dict = self._get_param_desc()

        # Root datainfo or memberinfo for nested datatypes
        self.datainfo: dict = deep_get(
            self._param_description["datainfo"], self.path.get_memberinfo_path()
        )

        self.readonly = self._param_description.get("readonly")

        self.SECoPdtype_str: str
        self.SECoPdtype_obj: DataType = self._param_description["datatype"]

        self.SECoP_type_info: SECoPdtype = SECoPdtype(self.SECoPdtype_obj)

        if self.SECoP_type_info.max_depth > MAX_DEPTH:
            warnings.warn(
                f"The datatype of parameter '{path._accessible_name}' has a maximum "
                f"depth of {self.SECoP_type_info.max_depth}. Tiled & Databroker only "
                f"support a Depth upto {MAX_DEPTH} "
                f"dtype_descr: {self.SECoP_type_info.dtype_descr}"
            )

        self.describe_dict: dict = {}

        self.source_name = (
            secclient.uri
            + ":"
            + secclient.nodename
            + ":"
            + self.path._module_name
            + ":"
            + self.path._accessible_name
        )

        # SECoP metadata is static and can only change when connection is reset
        self.describe_dict = {}

        self.describe_dict["source"] = self.source_name

        # add gathered keys from SECoPdtype:
        self.describe_dict.update(self.SECoP_type_info.get_datakey())

        for property_name, prop_val in self._param_description.items():
            # skip datainfo (treated seperately)
            if property_name == "datainfo" or property_name == "datatype":
                continue
            self.describe_dict[property_name] = prop_val

        for property_name, prop_val in self.datainfo.items():
            if property_name == "type":
                property_name = "SECoP_dtype"
            if property_name == "unit":
                property_name = "units"
            self.describe_dict[property_name] = prop_val

        super().__init__(datatype=self.SECoP_type_info.np_datatype)

    def source(self, name: str, read: bool) -> str:
        return self.source_name

    async def connect(self, timeout: float):
        pass

    async def put(self, value: Any | None, wait=True):
        # convert to frappy compatible Format
        secop_val = self.SECoP_type_info.val2secop(value)

        # frappy client has no ability to just send a secop message without
        # waiting for a reply
        await asyncio.wait_for(
            self._secclient.set_parameter(**self.get_param_path(), value=secop_val),
            timeout=None,
        )

    async def get_datakey(self, source: str) -> DataKey:
        """Metadata like source, dtype, shape, precision, units"""

        if self.SECoP_type_info._is_composite or isinstance(
            self.SECoPdtype_obj, ArrayOf
        ):
            # getlast cached value
            dataset = await self._secclient.get_parameter(
                **self.get_param_path(), trycache=True
            )

            # this ensures the datakey is updated to the latest cached value
            SECoPReading(entry=dataset, secop_dt=self.SECoP_type_info)
            self.describe_dict.update(self.SECoP_type_info.get_datakey())

        return self.describe_dict

    async def get_reading(self) -> Reading:
        dataset = await self._secclient.get_parameter(
            **self.get_param_path(), trycache=False
        )

        sec_reading = SECoPReading(entry=dataset, secop_dt=self.SECoP_type_info)

        return sec_reading.get_reading()

    async def get_value(self) -> T:
        dataset: Reading = await self.get_reading()

        return dataset["value"]  # type: ignore

    async def get_setpoint(self) -> T:
        return await self.get_value()

    def set_callback(self, callback: Callback[T] | None) -> None:
        def awaitify(sync_func):
            """Wrap a synchronous callable to allow ``await``'ing it"""

            @wraps(sync_func)
            async def async_func(*args, **kwargs):
                return sync_func(*args, **kwargs)

            return async_func

        def updateItem(module, parameter, entry: CacheItem):  # noqa: N802
            data = SECoPReading(secop_dt=self.SECoP_type_info, entry=entry)
            async_callback = awaitify(callback)

            asyncio.run_coroutine_threadsafe(
                async_callback(reading=data.get_reading()),
                self._secclient.loop,
            )

        if callback is not None:
            self._secclient.register_callback(self.get_path_tuple(), updateItem)
        else:
            self._secclient.unregister_callback(self.get_path_tuple(), updateItem)

    def _get_param_desc(self) -> dict:
        return deep_get(self._secclient.modules, self.path.get_param_desc_path())

    def get_param_path(self):
        return self.path.get_param_path()

    def get_path_tuple(self):
        return self.path.get_path_tuple()

    def get_unit(self):
        return self.describe_dict.get("units", None)

    def is_number(self) -> bool:
        if (
            self.describe_dict["dtype"] == "number"
            or self.describe_dict["dtype"] == "integer"
        ):
            return True

        return False


class PropertyBackend(SignalBackend):
    """Readonly backend for static SECoP Properties of Nodes/Modules"""

    def __init__(
        self, prop_key: str, property_dict: Dict[str, T], secclient: AsyncFrappyClient
    ) -> None:
        """Initializes PropertyBackend

        :param prop_key: Name of Property
        :type prop_key: str
        :param propertyDict: Dicitonary containing all properties of Node/Module
        :type propertyDict: Dict[str, T]
        :param secclient: SECoP client providing communication to the SEC Node
        :type secclient: AsyncFrappyClient
        """
        # secclient

        self._property_dict = property_dict
        self._prop_key = prop_key
        self._prop_value = self._property_dict[self._prop_key]
        self.SECoPdtype_obj: DataType = secop_dtype_obj_from_json(self._prop_value)
        self.SECoP_type_info: SECoPdtype = SECoPdtype(self.SECoPdtype_obj)

        if self.SECoP_type_info.max_depth > MAX_DEPTH:
            warnings.warn(
                f"The datatype of parameter '{prop_key}' has a maximum"
                f"depth of {self.SECoP_type_info.max_depth}. Tiled & Databroker only"
                f"support a Depth upto {MAX_DEPTH}"
                f"dtype_descr: {self.SECoP_type_info.dtype_descr}"
            )

        # SECoP metadata is static and can only change when connection is reset
        self.describe_dict = {}
        self.source_name = prop_key
        self.describe_dict["source"] = self.source_name

        # add gathered keys from SECoPdtype:
        self.describe_dict.update(self.SECoP_type_info.get_datakey())

        self._secclient: AsyncFrappyClient = secclient
        # TODO full property path

        super().__init__(datatype=self.SECoP_type_info.np_datatype)

    def source(self, name: str, read: bool) -> str:
        return str(self.source_name)

    async def connect(self, timeout: float):
        """Connect to underlying hardware"""
        pass

    async def put(self, value: Optional[T], wait=True):
        """Put a value to the PV, if wait then wait for completion for up to timeout"""
        # Properties are readonly
        pass

    async def get_datakey(self, source: str) -> DataKey:
        """Metadata like source, dtype, shape, precision, units"""
        return self.describe_dict

    async def get_reading(self) -> Reading:
        dataset = CacheItem(
            value=self._prop_value, timestamp=self._secclient.conn_timestamp
        )

        sec_reading = SECoPReading(entry=dataset, secop_dt=self.SECoP_type_info)

        return sec_reading.get_reading()

    async def get_value(self) -> T:
        dataset: Reading = await self.get_reading()

        return dataset["value"]  # type: ignore

    async def get_setpoint(self) -> T:
        return await self.get_value()

    def set_callback(self, callback: Callback[T] | None) -> None:
        pass


def secop_dtype_obj_from_json(prop_val):
    if isinstance(prop_val, str):
        return StringType()

    if isinstance(prop_val, (int, float)):
        return FloatRange()

    if isinstance(prop_val, bool):
        return BoolType()

    if isinstance(prop_val, dict):  # SECoP Structs/tuples --> numpy ndarray
        members = {}
        for key, elem in prop_val.items():
            members[key] = secop_dtype_obj_from_json(elem)

        return StructOf(**members)

    if isinstance(prop_val, list):
        # empty list, cannot infer proper type
        if not prop_val:
            return ArrayOf(FloatRange())
        # check if all elements have same Type:
        if all(isinstance(elem, type(prop_val[0])) for elem in prop_val):
            members = secop_dtype_obj_from_json(prop_val[0])
            return ArrayOf(members)
        else:
            members = []  # type: ignore
            for elem in prop_val:
                members.append(secop_dtype_obj_from_json(elem))  # type: ignore
            return TupleOf(*members)

    raise Exception(
        f"""unsupported datatype in Property:  {str(prop_val.__class__.__name__)}\n
        propval: {prop_val}"""
    )



# mypy: disable-error-code="attr-defined"
from secop_ophyd.util import Path, deep_get


def test_path_nested_tuples():
    path = Path(module_name="mod", parameter_name="param")

    assert path.get_signal_name() == "param"

    path = path.append(1)
    assert path.get_signal_name() == "param-1"

    path = path.append(2)

    assert path.get_signal_name() == "param-1-2"

    path = path.append(3)
    assert path.get_signal_name() == "param-1-2-3"

    path = path.append("new_param")
    assert path.get_signal_name() == "new_param"


def test_path_memberinfo(nested_param_description):
    path = Path(module_name="mod", parameter_name="_nested_struct")

    param_desc = nested_param_description["_nested_struct"]
    datainfo: dict = deep_get(param_desc["datainfo"], path.get_memberinfo_path())

    assert datainfo == param_desc["datainfo"]

    path_struct = path.append("pos_struct")

    datainfo = deep_get(param_desc["datainfo"], path_struct.get_memberinfo_path())

    assert datainfo == param_desc["datainfo"]["members"]["pos_struct"]

    path_tupl = path.append("tupl")

    datainfo = deep_get(param_desc["datainfo"], path_tupl.get_memberinfo_path())

    assert datainfo == param_desc["datainfo"]["members"]["tupl"]

    path_tupl_2 = path_tupl.append(1)

    assert path_tupl_2.get_memberinfo_path() == ["members", "tupl", "members", 1]

    datainfo = deep_get(param_desc["datainfo"], path_tupl_2.get_memberinfo_path())

    assert datainfo == param_desc["datainfo"]["members"]["tupl"]["members"][1]


def test_path_insert(struct_val: dict):
    path = Path(module_name="mod", parameter_name="_nested_struct")

    path_struct_mem = path.append("pos_struct")
    path_tupl_mem = path.append("tupl")

    path_struct_mem = path_struct_mem.append("x")
    path_tupl_mem = path_tupl_mem.append(1)

    new_x_val = 10.0
    new_tuple_val = 15.0

    new_dic = path_struct_mem.insert_val(struct_val, new_x_val)

    assert deep_get(struct_val, path_struct_mem._dev_path) == new_x_val
    assert deep_get(new_dic, path_struct_mem._dev_path) == new_x_val

    new_dic_tup = path_tupl_mem.insert_val(struct_val, new_tuple_val)

    assert deep_get(struct_val, path_tupl_mem._dev_path) == new_tuple_val
    assert deep_get(new_dic_tup, path_tupl_mem._dev_path) == new_tuple_val

    print(struct_val)



# mypy: disable-error-code="attr-defined"
import asyncio

import numpy as np
import xprocess
from ophyd_async.core import AsyncStatus, SignalR, observe_value
from xprocess import ProcessStarter, XProcessInfo

from secop_ophyd.SECoPDevices import SECoPMoveableDevice, SECoPNodeDevice


async def test_node_structure(cryo_sim, cryo_node_internal_loop: SECoPNodeDevice):
    assert isinstance(cryo_node_internal_loop, SECoPNodeDevice)
    await cryo_node_internal_loop.disconnect_async()


async def test_node_read(cryo_sim, cryo_node_internal_loop: SECoPNodeDevice):
    # Node device should return the readbacks of the read signals of the child devices
    val_read = await cryo_node_internal_loop.read()
    assert val_read != {}
    await cryo_node_internal_loop.disconnect_async()


async def test_node_describe(cryo_sim, cryo_node_internal_loop: SECoPNodeDevice):
    # Node device should return the descriptions of the read signals of the child
    # devices
    val_desc = await cryo_node_internal_loop.describe()
    assert val_desc != {}
    await cryo_node_internal_loop.disconnect_async()


async def test_node_module_describe(cryo_sim, cryo_node_internal_loop: SECoPNodeDevice):

    val_desc = await cryo_node_internal_loop.cryo.describe_configuration()
    conf = await cryo_node_internal_loop.cryo.read_configuration()

    assert val_desc != {}
    assert conf != {}
    await cryo_node_internal_loop.disconnect_async()


async def test_dev_read(cryo_sim, cryo_node_internal_loop: SECoPNodeDevice):

    cryo_dev: SECoPMoveableDevice = cryo_node_internal_loop.cryo
    cryo_val = await cryo_dev.read()
    val_name = cryo_dev.value.name

    read_val = cryo_val[val_name].get("value")

    assert read_val is not None
    assert read_val > 5
    await cryo_node_internal_loop.disconnect_async()


async def test_signal_read(cryo_sim, cryo_node_internal_loop: SECoPNodeDevice):

    cryo_dev: SECoPMoveableDevice = cryo_node_internal_loop.cryo

    p = await cryo_dev.p.get_value(cached=False)
    assert p == 40.0

    await cryo_node_internal_loop.disconnect_async()


async def test_signal_read_cached(cryo_sim, cryo_node_internal_loop: SECoPNodeDevice):

    cryo_dev: SECoPMoveableDevice = cryo_node_internal_loop.cryo

    p = await cryo_dev.p.get_value(cached=True)

    assert p == 40.0

    await cryo_node_internal_loop.disconnect_async()


async def test_signal_stage_unstage_read_cached(
    cryo_sim, cryo_node_internal_loop: SECoPNodeDevice
):

    cryo_dev: SECoPMoveableDevice = cryo_node_internal_loop.cryo

    await cryo_dev.value.stage()

    await asyncio.sleep(1)

    await cryo_dev.value.unstage()

    await asyncio.sleep(1)

    p = await cryo_dev.p.get_value(cached=True)

    assert p == 40.0

    await cryo_node_internal_loop.disconnect_async()


async def test_status(cryo_sim, cryo_node_internal_loop: SECoPNodeDevice):

    cryo_dev: SECoPMoveableDevice = cryo_node_internal_loop.cryo
    status: SignalR = cryo_dev.status

    async for current_stat in observe_value(status):
        assert current_stat["f0"] == 100

        if current_stat["f0"] == 100:
            break


async def test_trigger(cryo_sim, cryo_node_internal_loop: SECoPNodeDevice):
    cryo_dev: SECoPMoveableDevice = cryo_node_internal_loop.cryo

    val_old = await cryo_dev.read()

    val_cached = await cryo_dev.read()

    assert (
        val_old[cryo_dev.value.name]["value"]
        == val_cached[cryo_dev.value.name]["value"]
    )

    stat: AsyncStatus = cryo_dev.trigger()

    await stat

    val_new = await cryo_dev.read()

    assert (
        val_old[cryo_dev.value.name]["value"] != val_new[cryo_dev.value.name]["value"]
    )

    await cryo_node_internal_loop.disconnect_async()


# async def test_node_read_config(cryo_sim,cryo_node:SECoP_Node_Device):
#    # Node device has no read value, it has to return an empty dict
#    val_desc = await cryo_node.read_configuration()
#    assert  val_desc == {}


async def test_node_drive(cryo_sim, cryo_node_internal_loop: SECoPNodeDevice):
    cryo_dev: SECoPMoveableDevice = cryo_node_internal_loop.cryo

    target_old = await cryo_dev.target.read()

    new_target = 11.0

    old_target = target_old[cryo_dev.target.name].get("value")

    assert old_target != new_target

    stat = cryo_dev.set(new_target=new_target)

    # assert new_target == conf_new.get('target').get('value')

    await stat

    reading = await cryo_dev.value.get_value()

    assert np.isclose(reading, new_target, atol=0.2)

    await cryo_node_internal_loop.disconnect_async()


async def test_node_reconn(
    cryo_sim: xprocess, cryo_node_internal_loop: SECoPNodeDevice, frappy_env
):

    class Starter(ProcessStarter):
        # startup pattern
        pattern = ".*: startup done with interface(s)*"
        timeout = 5
        # command to start process
        env = frappy_env["env"]  # Use the environment variables from the fixture
        args = [
            "frappy-server",
            "-c",
            "cryo_cfg.py",
            "cryo",
        ]

    # cryo_dev: SECoPMoveableDevice = cryo_node_internal_loop.cryo

    old_conf = await cryo_node_internal_loop.read_configuration()

    # disconnect sec-node
    # clean up whole process tree afterwards

    cryo_info: XProcessInfo = cryo_sim.getinfo("cryo_sim")

    cryo_info.terminate(timeout=3)

    await asyncio.sleep(1)

    cryo_sim.ensure("cryo_sim", Starter)

    await asyncio.sleep(2)

    new_conf = await cryo_node_internal_loop.read_configuration()

    fw_old = old_conf[cryo_node_internal_loop.firmware.name]
    fw_new = new_conf[cryo_node_internal_loop.firmware.name]

    assert fw_new["timestamp"] > fw_old["timestamp"]

    await cryo_node_internal_loop.disconnect_async()



#####################################################################
# Python version of frappy config
#####################################################################
import os
import sys

# Get project root from environment variable
project_root = os.environ.get('FRAPPY_PROJECT_ROOT')

# If project_root is provided and it's a valid directory, add it to sys.path
if project_root and os.path.isdir(project_root):
    if project_root not in sys.path:
        sys.path.insert(0, project_root)


Node(
    "ophyd_secop.frappy.demo",
    "short description\n\n"
    "This is a very long description providing all the gory details "
    "about the stuff we are describing.",
    "tcp://10771",
    more="blub",
)

Mod(
    "str_test",
    "frappy_modules.ophyd_secop_test_modules.Test_Mod_str",
    "test module for reading String signals in Bluesky RE",
    group="test",
    value="blah",
)


Mod(
    "struct_of_arrays",
    "frappy_modules.ophyd_secop_test_modules.Test_Struct_of_arrays",
    "module for testing struct of arrays",
    value={
        "ints": [1, 2, 3, 4, 5],
        "strings": ["aaaa", "aaaaa", "aaaa", "aaa", "aa"],
        "floats": [1.1, 2.2, 4.3, 6.4, 7.5],
    },
    writable_strct_of_arr={
        "ints": [1, 2, 3, 4, 5],
        "strings": ["aaaa", "aaaaa", "aaaa", "aaa", "aa"],
        "floats": [1.1, 2.2, 4.3, 6.4, 7.5],
    },
)


Mod(
    "nd_arr",
    "frappy_modules.ophyd_secop_test_modules.Test_ND_arrays",
    "module for testing multidimensional arrays",
    value=[
        [1, 1, 1, 1, 1],
        [1, 1, 1, 1, 1],
        [1, 1, 1, 1, 1],
        [1, 1, 1, 1, 1],
        [1, 1, 1, 1, 1],
    ],
    arr3d=[[[0]]],
)

Mod(
    "primitive_arrays",
    "frappy_modules.ophyd_secop_test_modules.OPHYD_test_primitive_arrays",
    "simulated hardware for testing handling of arays of primitives",
    value=10.3,
)


Mod(
    "ophy_struct",
    "frappy_modules.ophyd_secop_test_modules.OPYD_test_struct",
    "simulated hardware for testing ophyd struct behavior" "also tuples",
    group="very important/stuff",
    # jitter=0.1,
    p_start=5,
    looptime=1,
    value={"x": 10, "y": 10, "z": 10, "color": "blue"},
    target={"x": 10, "y": 10, "z": 10, "color": "blue"},
    nested_struct={
        "number": 5,
        "string": "blablalbla",
        "tupl": [1, 1, 1],
        "pos_struct": {"x": 5, "y": 10, "z": 15, "col": "green", "enum": 1},
    },
    tuple_param=(5, 5, 5, "green"),
    # ramp=6,
    # maxpower=20.0,
    # heater=4.1,
    # mode='pid',
    # tolerance=0.1,
    # window=30,
    # timeout=900,
    # p = Param(40, unit='%/K'),  # in case 'default' is the first arg, we can omit 'default='
    # i = 10,
    # d = 2,
    # pid = Group('p', 'i', 'd'),
    # pollinterval = Param(export=False),
    # value = Param(unit = 'K', test = 'customized value'),
)



#####################################################################
# Python version of frappy config
#####################################################################
import os
import sys

# Get project root from environment variable
project_root = os.environ.get('FRAPPY_PROJECT_ROOT')

# If project_root is provided and it's a valid directory, add it to sys.path
if project_root and os.path.isdir(project_root):
    if project_root not in sys.path:
        sys.path.insert(0, project_root)



Node(
    "cryo_7.frappy.demo",
    "short description\n\n"
    "This is a very long description providing all the gory details "
    "about the stuff we are describing.",
    "tcp://10769",
    more="blub",
)

Mod(
    "cryo",
    "frappy_modules.cryo.Cryostat",
    "A simulated cc cryostat with heat-load, specific heat for the sample and a "
    "temperature dependent heat-link between sample and regulation.",
    group="very important/stuff",
    jitter=0.1,
    T_start=10.0,
    target=10.0,
    looptime=1,
    ramp=6,
    maxpower=20.0,
    heater=4.1,
    mode="pid",
    tolerance=0.1,
    window=30,
    timeout=900,
    p=Param(
        40, unit="%/K"
    ),  # in case 'default' is the first arg, we can omit 'default='
    i=10,
    d=2,
    pid=Group("p", "i", "d"),
    pollinterval=Param(export=False),
    value=Param(unit="K", test="customized value"),
)



import asyncio
import inspect
import re
import threading
import time as ttime
from types import MethodType
from typing import Any, Dict, Iterator, Optional, Type

import bluesky.plan_stubs as bps
from bluesky.protocols import (
    Descriptor,
    Flyable,
    Locatable,
    Location,
    PartialEvent,
    Reading,
    Stoppable,
    Subscribable,
    SyncOrAsync,
    Triggerable,
)
from frappy.client import Logger
from frappy.datatypes import (
    ArrayOf,
    BLOBType,
    BoolType,
    CommandType,
    FloatRange,
    IntRange,
    ScaledInteger,
    StringType,
    StructOf,
    TupleOf,
)
from ophyd_async.core import (
    AsyncStatus,
    SignalR,
    SignalRW,
    SignalX,
    StandardReadable,
    StandardReadableFormat,
    observe_value,
)
from ophyd_async.core._utils import Callback

from secop_ophyd.AsyncFrappyClient import AsyncFrappyClient
from secop_ophyd.GenNodeCode import GenNodeCode, Method
from secop_ophyd.propertykeys import DATAINFO, EQUIPMENT_ID, INTERFACE_CLASSES
from secop_ophyd.SECoPSignal import (
    LocalBackend,
    PropertyBackend,
    SECoPParamBackend,
    SECoPXBackend,
)
from secop_ophyd.util import Access, Path, Role, get_access_level

# Predefined Status Codes
DISABLED = 0
IDLE = 100
STANDBY = 130
PREPARED = 150
WARN = 200
WARN_STANDBY = 230
WARN_PREPARED = 250
NSTABLE = 270  # not in SECoP standard (yet)
BUSY = 300
DISABLING = 310
INITIALIZING = 320
PREPARING = 340
STARTING = 360
RAMPING = 370
STABILIZING = 380
FINALIZING = 390
ERROR = 400
ERROR_STANDBY = 430
ERROR_PREPARED = 450
UNKNOWN = 401  # not in SECoP standard (yet)


READABLE_PARAMS = ["value", "target"]


def clean_identifier(anystring):
    return str(re.sub(r"\W+|^(?=\d)", "_", anystring))


def get_config_attrs(parameters):
    parameters_cfg = parameters.copy()
    parameters_cfg.pop("target", None)
    parameters_cfg.pop("value", None)
    return parameters_cfg


class SECoPCMDDevice(StandardReadable, Flyable, Triggerable):
    """
    Command devices that have Signals for command args, return values and a signal
    for triggering command execution (SignalX). They themselves are triggerable.

    Once the CMD Device is triggered, the command args are retrieved from the 'argument'
    Signal. The command message is sent to the SEC Node and the return value is written
    to 'result' signal.

    """

    def __init__(self, path: Path, secclient: AsyncFrappyClient):
        """Initialize the CMD Device

        :param path: Path to the command in the secclient module dict
        :type path: Path
        :param secclient: SECoP client providing communication to the SEC Node
        :type secclient: AsyncFrappyClient
        """
        dev_name: str = path.get_signal_name() + "_CMD"

        self._secclient: AsyncFrappyClient = secclient

        cmd_props = secclient.modules[path._module_name]["commands"][
            path._accessible_name
        ]  # noqa: E501
        cmd_datatype: CommandType = cmd_props["datatype"]
        datainfo = cmd_props[DATAINFO]

        self.description: str = cmd_props["description"]
        self.arg_dtype = cmd_datatype.argument
        self.res_dtype = cmd_datatype.result

        self.argument: SignalRW | None
        self.result: SignalR | None

        # result signals
        read = []
        # argument signals
        config = []

        self._start_time: float
        self.commandx: SignalX

        self.wait_idle: bool = False

        with self.add_children_as_readables(
            format=StandardReadableFormat.CONFIG_SIGNAL
        ):
            # Argument Signals (config Signals, can also be read)
            arg_path = path.append("argument")
            if self.arg_dtype is None:
                self.argument = None
            else:
                arg_backend = LocalBackend(
                    path=arg_path,
                    secop_dtype_obj=self.arg_dtype,
                    sig_datainfo=datainfo["argument"],
                )
                self.argument = SignalRW(arg_backend)
                config.append(self.argument)

            # Result Signals  (read Signals)
            res_path = path.append("result")

            if self.res_dtype is None:
                self.result = None
            else:
                res_backend = LocalBackend(
                    path=res_path,
                    secop_dtype_obj=self.res_dtype,
                    sig_datainfo=datainfo["result"],
                )
                self.result = SignalRW(res_backend)
                read.append(self.argument)

            argument = None
            result = None
            if isinstance(self.argument, SignalR):
                argument = self.argument._connector.backend

            if isinstance(self.result, SignalR):
                result = self.result._connector.backend

            # SignalX (signal that triggers execution of the Command)
            exec_backend = SECoPXBackend(
                path=path,
                secclient=secclient,
                argument=argument,  # type: ignore
                result=result,  # type: ignore
            )

        self.commandx = SignalX(exec_backend)

        super().__init__(name=dev_name)

    def trigger(self) -> AsyncStatus:
        """Triggers the SECoPCMDDevice and sends command message to SEC Node.
        Command argument is taken form 'argument' Signal, and return value is
        written in the 'return' Signal

        :return: A Status object, that is marked Done once the answer from the
        SEC Node is received
        :rtype: AsyncStatus
        """
        coro = asyncio.wait_for(fut=self._exec_cmd(), timeout=None)
        return AsyncStatus(awaitable=coro)

    def kickoff(self) -> AsyncStatus:
        # trigger execution of secop command, wait until Device is Busy

        self._start_time = ttime.time()
        coro = asyncio.wait_for(fut=asyncio.sleep(1), timeout=None)
        return AsyncStatus(coro)

    async def _exec_cmd(self):
        stat = self.commandx.trigger()

        await stat

    def complete(self) -> AsyncStatus:
        coro = asyncio.wait_for(fut=self._exec_cmd(), timeout=None)
        return AsyncStatus(awaitable=coro)

    def collect(self) -> Iterator[PartialEvent]:
        yield dict(
            time=self._start_time, timestamps={self.name: []}, data={self.name: []}
        )

    async def describe_collect(self) -> SyncOrAsync[Dict[str, Dict[str, Descriptor]]]:
        return await self.describe()


class SECoPBaseDevice(StandardReadable):
    """Base Class for generating Opyd devices from SEC Node modules,
    objects of type SECoPBaseDevice are not supposed to be instanciated

    """

    def __init__(
        self, secclient: AsyncFrappyClient, module_name: str, role: Role = Role.USER
    ) -> None:
        """Initiate A SECoPBaseDevice

        :param secclient: SECoP client providing communication to the SEC Node
        :type secclient: AsyncFrappyClient
        """

        self._secclient: AsyncFrappyClient = secclient

        self.impl: str | None = None

        self.role: Role = role

        self._module = module_name
        module_desc = secclient.modules[module_name]
        self.plans: list[Method] = []
        self.mod_prop_devices: Dict[str, SignalR] = {}
        self.param_devices: Dict[str, Any] = {}

        # Add configuration Signal
        with self.add_children_as_readables(
            format=StandardReadableFormat.CONFIG_SIGNAL
        ):
            # generate Signals from Module Properties
            for property in module_desc["properties"]:
                propb = PropertyBackend(property, module_desc["properties"], secclient)
                if property == "implementation":
                    self.impl = module_desc["properties"]["implementation"]

                if property == "meaning":
                    continue

                setattr(self, property, SignalR(backend=propb))
                self.mod_prop_devices[property] = getattr(self, property)

            # generate Signals from Module parameters eiter r or rw
            for parameter, properties in module_desc["parameters"].items():
                if parameter in READABLE_PARAMS:
                    continue
                # generate new root path
                param_path = Path(parameter_name=parameter, module_name=module_name)

                # readonly propertyns to plans and plan stubs.
                readonly: bool = properties.get("readonly", None)

                # Normal types + (struct and tuple as JSON object Strings)
                self._signal_from_parameter(
                    path=param_path,
                    sig_name=parameter,
                    readonly=readonly,
                )
                self.param_devices[parameter] = getattr(self, parameter)

        # Add readables Signals
        with self.add_children_as_readables(
            format=StandardReadableFormat.HINTED_SIGNAL
        ):
            for parameter in READABLE_PARAMS:
                if parameter not in module_desc["parameters"].keys():
                    continue
                properties = module_desc["parameters"][parameter]

                # generate new root path
                param_path = Path(parameter_name=parameter, module_name=module_name)

                # readonly propertyns to plans and plan stubs.
                readonly = properties.get("readonly", None)

                # Normal types + (struct and tuple as JSON object Strings)
                self._signal_from_parameter(
                    path=param_path,
                    sig_name=parameter,
                    readonly=readonly,
                )
                self.param_devices[parameter] = getattr(self, parameter)

        # Initialize Command Devices
        for command, properties in module_desc["commands"].items():
            # generate new root path
            cmd_path = Path(parameter_name=command, module_name=module_name)
            cmd_dev_name = command + "_CMD"
            setattr(
                self,
                cmd_dev_name,
                SECoPCMDDevice(path=cmd_path, secclient=secclient),
            )

            cmd_dev: SECoPCMDDevice = getattr(self, cmd_dev_name)
            # Add Bluesky Plan Methods

            # Stop is already an ophyd native operation
            if command == "stop":
                continue

            cmd_plan = self.generate_cmd_plan(
                cmd_dev, cmd_dev.arg_dtype, cmd_dev.res_dtype
            )

            setattr(self, command, MethodType(cmd_plan, self))

            description: str = ""
            description += f"{cmd_dev.description}\n"
            description += f"       argument: {str(cmd_dev.arg_dtype)}\n"
            description += f"       result: {str(cmd_dev.res_dtype)}"

            plan = Method(
                cmd_name=command,
                description=description,
                cmd_sign=inspect.signature(getattr(self, command)),
            )

            self.plans.append(plan)

        self.set_name(module_name)

    def generate_cmd_plan(
        self,
        cmd_dev: SECoPCMDDevice,
        argument_type: Type | None = None,
        return_type: Type | None = None,
    ):

        def command_plan_no_arg(self, wait_for_idle: bool = False):
            # Trigger the Command device, meaning that the command gets sent to the
            # SEC Node
            yield from bps.trigger(cmd_dev, wait=True)

            if wait_for_idle:

                def wait_for_idle_factory():
                    return self.wait_for_idle()

                yield from bps.wait_for([wait_for_idle_factory])

            if (
                return_type is not None
                and isinstance(cmd_dev.result, SignalR)
                and isinstance(cmd_dev.result._connector.backend, LocalBackend)
            ):

                return cmd_dev.result._connector.backend.reading.get_value()

        def command_plan(self, arg, wait_for_idle: bool = False):
            # TODO  Type checking

            if arg is not None:
                yield from bps.abs_set(cmd_dev.argument, arg)

            # Trigger the Command device, meaning that the command gets sent to the
            # SEC Node
            yield from bps.trigger(cmd_dev, wait=True)

            if wait_for_idle:

                def wait_for_idle_factory():
                    return self.wait_for_idle()

                yield from bps.wait_for([wait_for_idle_factory])

            if (
                return_type is not None
                and isinstance(cmd_dev.result, SignalR)
                and isinstance(cmd_dev.result._connector.backend, LocalBackend)
            ):

                return cmd_dev.result._connector.backend.reading.get_value()

        cmd_meth = command_plan_no_arg if argument_type is None else command_plan

        anno_dict = cmd_meth.__annotations__

        dtype_mapping = {
            StructOf: dict[str, Any],
            ArrayOf: list[Any],
            TupleOf: tuple[Any],
            BLOBType: str,
            BoolType: bool,
            FloatRange: float,
            IntRange: int,
            ScaledInteger: int,
            StringType: str,
        }

        if return_type is not None:
            anno_dict["return"] = dtype_mapping[return_type.__class__]
        if argument_type is not None:
            anno_dict["arg"] = dtype_mapping[argument_type.__class__]

        return cmd_meth

    def _signal_from_parameter(self, path: Path, sig_name: str, readonly: bool):
        """Generates an Ophyd Signal from a Module Parameter

        :param path: Path to the Parameter in the secclient module dict
        :type path: Path
        :param sig_name: Name of the new Signal
        :type sig_name: str
        :param readonly: Signal is R or RW
        :type readonly: bool
        """
        # Normal types + (struct and tuple as JSON object Strings)
        paramb = SECoPParamBackend(path=path, secclient=self._secclient)

        # construct signal
        if readonly:
            setattr(self, sig_name, SignalR(paramb))
        else:
            setattr(self, sig_name, SignalRW(paramb))

        def noop(val):
            pass

        sig: SignalR = getattr(self, sig_name)
        sig.subscribe_value(noop)


class SECoPCommunicatorDevice(SECoPBaseDevice):

    def __init__(
        self, secclient: AsyncFrappyClient, module_name: str, role: Role = Role.USER
    ):
        """Initializes the SECoPCommunicatorDevice

        :param secclient: SECoP client providing communication to the SEC Node
        :type secclient: AsyncFrappyClient
        :param module_name: Name of the SEC Node module that is represented by
            this device
        :type module_name: str"""

        super().__init__(secclient=secclient, module_name=module_name, role=role)


class SECoPReadableDevice(SECoPCommunicatorDevice, Triggerable, Subscribable):
    """
    Standard readable SECoP device, corresponding to a SECoP module with the
    interface class "Readable"
    """

    def __init__(
        self, secclient: AsyncFrappyClient, module_name: str, role: Role = Role.USER
    ):
        """Initializes the SECoPReadableDevice

        :param secclient: SECoP client providing communication to the SEC Node
        :type secclient: AsyncFrappyClient
        :param module_name: Name of the SEC Node module that is represented by
            this device
        :type module_name: str
        """

        self.value: SignalR
        self.status: SignalR

        super().__init__(secclient=secclient, module_name=module_name, role=role)

        if not hasattr(self, "value"):
            raise AttributeError(
                "Attribute 'value' has not been assigned,"
                + "but is needed for Readable interface class"
            )

        if not hasattr(self, "status"):
            raise AttributeError(
                "Attribute 'status' has not been assigned,"
                + "but is needed for Readable interface class"
            )

    async def wait_for_idle(self):
        """asynchronously waits until module is IDLE again. this is helpful,
        for running commands that are not done immediately
        """
        if self.status is None:
            raise Exception("status Signal not initialized")

        # force reading of fresh status from device
        await self.status.read(False)

        async for current_stat in observe_value(self.status):
            # status is has type Tuple and is therefore transported as
            # structured Numpy array ('f0':statuscode;'f1':status Message)

            stat_code = current_stat["f0"]

            # Module is in IDLE/WARN state
            if IDLE <= stat_code < BUSY:
                break

            if hasattr(self, "_stopped"):
                if self._stopped is True:
                    break

            # Error State or DISABLED
            if hasattr(self, "_success"):
                if stat_code >= ERROR or stat_code < IDLE:
                    self._success = False
                    break

    # TODO add timeout
    def observe_status_change(self, monitored_status_code: int):
        async def switch_from_status_inner():
            async for current_stat in observe_value(self.status):
                # status is has type Tuple and is therefore transported as
                # structured Numpy array ('f0':statuscode;'f1':status Message)

                stat_code = current_stat["f0"]

                if monitored_status_code != stat_code:
                    break

        def switch_from_status_factory():
            return switch_from_status_inner()

        yield from bps.wait_for([switch_from_status_factory])

    def trigger(self) -> AsyncStatus:
        return AsyncStatus(awaitable=self.value.read(cached=False))

    def subscribe(self, function: Callback[dict[str, Reading]]) -> None:
        """Subscribe to updates in the reading"""
        self.value.subscribe(function=function)

    def clear_sub(self, function: Callback) -> None:
        """Remove a subscription."""
        self.value.clear_sub(function=function)


class SECoPTriggerableDevice(SECoPReadableDevice, Stoppable):
    """
    Standard triggerable SECoP device, corresponding to a SECoP module with the0s
    interface class "Triggerable"
    """

    def __init__(
        self, secclient: AsyncFrappyClient, module_name: str, role: Role = Role.USER
    ):
        """Initialize SECoPTriggerableDevice

        :param secclient: SECoP client providing communication to the SEC Node
        :type secclient: AsyncFrappyClient
        :param module_name: ame of the SEC Node module that is represented by
            this device
        :type module_name: str
        """

        self.go_CMD: SECoPCMDDevice

        self._success = True
        self._stopped = False

        super().__init__(secclient, module_name, role)

    async def __go_coro(self, wait_for_idle: bool):
        await self._secclient.exec_command(module=self._module, command="go")

        self._success = True
        self._stopped = False
        await asyncio.sleep(0.2)

        if wait_for_idle:
            await self.wait_for_idle()

    def wait_for_prepared(self):
        yield from self.observe_status_change(IDLE)
        yield from self.observe_status_change(PREPARING)

    def trigger(self) -> AsyncStatus:

        async def go_or_read_on_busy():
            module_status = await self.status.get_value(False)
            stat_code = module_status["f0"]

            if BUSY <= stat_code <= ERROR:
                return

            await self.__go_coro(True)

        return AsyncStatus(awaitable=go_or_read_on_busy())

    async def stop(self, success=True):
        """Calls stop command on the SEC Node module

        :param success:
            True: device is stopped as planned
            False: something has gone wrong
            (defaults to True)
        :type success: bool, optional
        """
        self._success = success

        await self._secclient.exec_command(self._module, "stop")
        self._stopped = True


class SECoPWritableDevice(SECoPReadableDevice):
    """Fast settable device target"""

    pass


class SECoPMoveableDevice(SECoPWritableDevice, Locatable, Stoppable):
    """
    Standard movable SECoP device, corresponding to a SECoP module with the
    interface class "Drivable"
    """

    def __init__(
        self, secclient: AsyncFrappyClient, module_name: str, role: Role = Role.USER
    ):
        """Initialize SECoPMovableDevice

        :param secclient: SECoP client providing communication to the SEC Node
        :type secclient: AsyncFrappyClient
        :param module_name: ame of the SEC Node module that is represented by
            this device
        :type module_name: str
        """

        self.target: SignalRW

        super().__init__(secclient, module_name, role)

        if not hasattr(self, "target"):
            raise AttributeError(
                "Attribute 'target' has not been assigned, "
                + "but is needed for 'Drivable' interface class!"
            )

        self._success = True
        self._stopped = False

    def set(self, new_target, timeout: Optional[float] = None) -> AsyncStatus:
        """Sends new target to SEC Nonde and waits until module is IDLE again

        :param new_target: new taget/setpoint for module
        :type new_target: _type_
        :param timeout: timeout for set operation, defaults to None
        :type timeout: Optional[float], optional
        :return: Asyncstatus that gets set to Done once module is IDLE again
        :rtype: AsyncStatus
        """
        coro = asyncio.wait_for(self._move(new_target), timeout=timeout)
        return AsyncStatus(coro)

    async def _move(self, new_target):
        self._success = True
        self._stopped = False
        await self.target.set(new_target, wait=False)

        # force reading of status from device
        await self.status.read(False)

        # observe status and wait until dvice is IDLE again
        async for current_stat in observe_value(self.status):
            stat_code = current_stat["f0"]

            if self._stopped is True:
                break

            # Error State or DISABLED
            if stat_code >= ERROR or stat_code < IDLE:
                self._success = False
                break

            # Module is in IDLE/WARN state
            if IDLE <= stat_code < BUSY:
                break

            # TODO other status transitions

        if not self._success:
            raise RuntimeError("Module was stopped")

    async def stop(self, success=True):
        """Calls stop command on the SEC Node module

        :param success:
            True: device is stopped as planned
            False: something has gone wrong
            (defaults to True)
        :type success: bool, optional
        """
        self._success = success

        await self._secclient.exec_command(self._module, "stop")
        self._stopped = True

    async def locate(self) -> Location:
        # return current location of the device (setpoint and readback).
        # Only locally cached values are returned
        location: Location = {
            "setpoint": await self.target.get_value(True),
            "readback": await self.value.get_value(True),
        }
        return location


class SECoPNodeDevice(StandardReadable):
    """
    Generates the root ophyd device from a Sec-node. Signals of this Device correspond
    to the Sec-node properties
    """

    def __init__(self, secclient: AsyncFrappyClient, role="USER"):
        """Initializes the node device and generates all node signals and subdevices
        corresponding to the SECoP-modules of the secnode

        :param secclient: SECoP client providing communication to the SEC Node
        :type secclient: AsyncFrappyClient
        """

        self.role: Role = Role[role]

        self.equipment_id: SignalR
        self.description: SignalR
        self.version: SignalR

        self._secclient: AsyncFrappyClient = secclient

        self._module_name: str = ""
        self._node_cls_name: str = ""
        self.mod_devices: Dict[str, SECoPReadableDevice] = {}
        self.node_prop_devices: Dict[str, SignalR] = {}

        self.genCode: GenNodeCode

        # Name is set to sec-node equipment_id
        name = self._secclient.properties[EQUIPMENT_ID].replace(".", "-")

        config = []

        with self.add_children_as_readables(
            format=StandardReadableFormat.CONFIG_SIGNAL
        ):
            for property in self._secclient.properties:
                propb = PropertyBackend(property, self._secclient.properties, secclient)
                setattr(self, property, SignalR(backend=propb))
                config.append(getattr(self, property))
                self.node_prop_devices[property] = getattr(self, property)

        with self.add_children_as_readables(format=StandardReadableFormat.CHILD):
            for module, module_desc in self._secclient.modules.items():

                secop_dev_class = self.class_from_interface(module_desc["properties"])

                # if access level is not enough 'secop_dev_class' is set to None
                # --> module is hidden
                if secop_dev_class is not None:
                    setattr(
                        self,
                        module,
                        secop_dev_class(self._secclient, module, self.role),
                    )
                    self.mod_devices[module] = getattr(self, module)

        # register secclient callbacks (these are useful if sec node description
        # changes after a reconnect)
        secclient.client.register_callback(
            None, self.descriptiveDataChange, self.nodeStateChange
        )

        super().__init__(name=name)

    @classmethod
    def create(cls, host: str, port: str, loop, log=Logger) -> "SECoPNodeDevice":

        secclient: AsyncFrappyClient

        if not loop.is_running():
            raise Exception("The provided Eventloop is not running")

        if loop._thread_id != threading.current_thread().ident:
            client_future = asyncio.run_coroutine_threadsafe(
                AsyncFrappyClient.create(host=host, port=port, loop=loop, log=log), loop
            )
            secclient = client_future.result()

            secclient.external = True

        else:
            raise Exception(
                "should be calles with an eventloop that is"
                "running in a seperate thread"
            )

        return SECoPNodeDevice(secclient=secclient)

    @classmethod
    async def create_async(
        cls, host: str, port: str, loop, log=Logger
    ) -> "SECoPNodeDevice":

        secclient: AsyncFrappyClient

        if not loop.is_running():
            raise Exception("The provided Eventloop is not running")

        if loop._thread_id == threading.current_thread().ident:
            secclient = await AsyncFrappyClient.create(
                host=host, port=port, loop=loop, log=log
            )

            secclient.external = False

        else:
            # Event loop is running in a different thread
            client_future = asyncio.run_coroutine_threadsafe(
                AsyncFrappyClient.create(host=host, port=port, loop=loop, log=log), loop
            )
            secclient = await asyncio.wrap_future(future=client_future)
            secclient.external = True

        return SECoPNodeDevice(secclient=secclient)

    def disconnect(self):
        """shuts down secclient, eventloop must be running in external thread"""
        if (
            self._secclient.loop._thread_id == threading.current_thread().ident
            and self._secclient.loop.is_running()
        ):
            raise Exception(
                "Eventloop must be running in external thread,"
                " try await node.disconnect_async()"
            )
        else:
            future = asyncio.run_coroutine_threadsafe(
                self._secclient.disconnect(True), self._secclient.loop
            )

            future.result(2)

    async def disconnect_async(self):
        """shuts down secclient using asyncio, eventloop can be running in same or
        external thread
        """
        if (
            self._secclient.loop._thread_id == threading.current_thread().ident
            and self._secclient.loop.is_running()
        ):
            await self._secclient.disconnect(True)
        else:
            disconn_future = asyncio.run_coroutine_threadsafe(
                self._secclient.disconnect(True), self._secclient.loop
            )
            await asyncio.wrap_future(future=disconn_future)

    def class_from_instance(self, path_to_module: str | None = None):
        """Dynamically generate python class file for the SECoP_Node_Device, this
        allows autocompletion in IDEs and eases working with the generated Ophyd
        devices
        """

        # parse genClass file if already present
        self.genCode = GenNodeCode(path=path_to_module, log=self._secclient.log)

        self.genCode.add_import(self.__module__, self.__class__.__name__)

        node_dict = self.__dict__

        # NodeClass Name
        self._node_cls_name = self.name.replace("-", "_").capitalize()

        node_bases = [self.__class__.__name__, "ABC"]

        node_class_attrs = []

        for attr_name, attr_value in node_dict.items():
            # Modules
            if isinstance(
                attr_value,
                (
                    SECoPBaseDevice,
                    SECoPCommunicatorDevice,
                    SECoPReadableDevice,
                    SECoPWritableDevice,
                    SECoPMoveableDevice,
                    SECoPTriggerableDevice,
                ),
            ):
                attr_type = type(attr_value)
                module = str(getattr(attr_type, "__module__", None))

                # add imports for module attributes
                self.genCode.add_import(module, attr_type.__name__)

                module_dict = attr_value.__dict__

                # modclass is baseclass of derived class
                mod_bases = [attr_value.__class__.__name__, "ABC"]

                module_class_attrs = []

                # Name for derived class
                module_class_name = attr_name
                if attr_value.impl is not None:
                    module_class_name = attr_value.impl.split(".").pop()

                # Module:Acessibles
                for module_attr_name, module_attr_value in module_dict.items():
                    if isinstance(
                        module_attr_value,
                        (SignalR, SignalX, SignalRW, SignalR, SECoPCMDDevice),
                    ):
                        # add imports for module attributes
                        self.genCode.add_import(
                            module_attr_value.__module__,
                            type(module_attr_value).__name__,
                        )

                        module_class_attrs.append(
                            (module_attr_name, type(module_attr_value).__name__)
                        )
                self.genCode.add_mod_class(
                    module_class_name, mod_bases, module_class_attrs, attr_value.plans
                )

                node_class_attrs.append((attr_name, module_class_name))

            # Poperty Signals
            if isinstance(attr_value, (SignalR)):
                self.genCode.add_import(
                    attr_value.__module__, type(attr_value).__name__
                )
                node_class_attrs.append((attr_name, attr_value.__class__.__name__))

        self.genCode.add_node_class(self._node_cls_name, node_bases, node_class_attrs)

        self.genCode.write_gen_node_class_file()

    def descriptiveDataChange(self, module, description):  # noqa: N802
        """called when the description has changed

        this callback is called on the node with module=None
        and on every changed module with module==<module name>

        :param module: module name of the module that has changes
        :type module: _type_
        :param description: new Node description string
        :type description: _type_
        """
        # TODO this functionality is untested and will probably break the generated
        # ophyd device since a changed module description would lead a newly
        # instanciated module object while references to the old one are broken
        # mitigation: alway call methods via:
        #
        # 'node_obj.module_obj.method()'

        self._secclient.conn_timestamp = ttime.time()

        if module is None:
            # Refresh signals that correspond to Node Properties
            config = []
            for property in self._secclient.properties:
                propb = PropertyBackend(
                    property, self._secclient.properties, self._secclient
                )

                setattr(self, property, SignalR(backend=propb))
                config.append(getattr(self, property))

            self.add_readables(config, format=StandardReadableFormat.CONFIG_SIGNAL)
        else:
            # Refresh changed modules
            module_desc = self._secclient.modules[module]
            secop_dev_class = self.class_from_interface(module_desc["properties"])

            setattr(self, module, secop_dev_class(self._secclient, module))

            # TODO what about removing Modules during disconn

    def nodeStateChange(self, online, state):  # noqa: N802
        """called when the state of the connection changes

        'online' is True when connected or reconnecting, False when disconnected
        or connecting 'state' is the connection state as a string
        """
        if state == "connected" and online is True:
            self._secclient.conn_timestamp = ttime.time()

    def class_from_interface(self, mod_properties: dict):
        module_access_level = Access.WRITE
        ophyd_class = None

        # check if acessmode is defined for the module
        if "accessmode" in mod_properties:
            module_access_level = get_access_level(
                self.role, mod_properties["accessmode"]
            )

        # infer highest level IF class
        module_interface_classes: dict = mod_properties[INTERFACE_CLASSES]
        for interface_class in IF_CLASSES.keys():
            if interface_class in module_interface_classes:
                ophyd_class = IF_CLASSES[interface_class]
                break

        # No predefined IF class was a match --> use base class (loose collection of
        # accessibles)
        if ophyd_class is None:
            ophyd_class = SECoPBaseDevice  # type: ignore

        # downgrade IF class if accesslevel is not sufficient
        match module_access_level:
            # Full access
            case Access.WRITE:
                return ophyd_class

            # demote to lower IF class
            case Access.READ:
                if ophyd_class == SECoPBaseDevice:
                    return SECoPBaseDevice
                if ophyd_class == SECoPCommunicatorDevice:
                    return SECoPCommunicatorDevice
                if ophyd_class == SECoPReadableDevice:
                    return SECoPReadableDevice
                if ophyd_class == SECoPWritableDevice:
                    return SECoPReadableDevice
                if ophyd_class == SECoPMoveableDevice:
                    return SECoPReadableDevice
                if ophyd_class == SECoPTriggerableDevice:
                    return SECoPReadableDevice

            # NO access --> Module is hidden
            case Access.NO_ACCESS:
                return None


IF_CLASSES = {
    "Triggerable": SECoPTriggerableDevice,
    "Drivable": SECoPMoveableDevice,
    "Writable": SECoPWritableDevice,
    "Readable": SECoPReadableDevice,
    "Communicator": SECoPCommunicatorDevice,
}

SECOP_TO_NEXUS_TYPE = {
    "double": "NX_FLOAT64",
    "int": "NX_INT64",
    "scaled": "NX_FLOAT64",
}


ALL_IF_CLASSES = set(IF_CLASSES.values())

# TODO
# FEATURES = {
#    'HasLimits': SecopHasLimits,
#    'HasOffset': SecopHasOffset,
# }



import random
import string
import time

import numpy as np
from frappy.core import Command, Drivable, IntRange, Parameter, Readable, StructOf
from frappy.datatypes import (
    ArrayOf,
    BoolType,
    EnumType,
    FloatRange,
    StatusType,
    StringType,
    TupleOf,
)
from frappy.errors import ImpossibleError
from frappy.lib import clamp, mkthread
from frappy.lib.enum import Enum

# test custom property (value.test can be changed in config file)


class Test_Mod_str(Readable):
    Status = Enum(Readable.Status)

    status = Parameter(datatype=StatusType(Status))

    value = Parameter("string read value", StringType())

    def get_random_string(self, length):
        # choose from all lowercase letter
        letters = string.ascii_lowercase
        return "".join(random.choice(letters) for i in range(length))

    def read_value(self):
        return self.get_random_string(10)

    def read_status(self):
        # instead of asking a 'Hardware' take the value from the simulation
        return self.status


class Test_ND_arrays(Readable):
    Status = Enum(Readable.Status)

    status = Parameter(datatype=StatusType(Status))

    value = Parameter(
        "2D integer Array",
        datatype=ArrayOf(
            members=ArrayOf(members=IntRange(min=0, max=100), minlen=5, maxlen=5),
            minlen=5,
            maxlen=5,
        ),
        readonly=True,
    )

    arr3d = Parameter(
        "3D integer Array",
        datatype=ArrayOf(
            ArrayOf(
                members=ArrayOf(
                    members=IntRange(min=0, max=100), minlen=5, maxlen=5
                ),
                minlen=5,
                maxlen=5,
            ),
            minlen=5,
            maxlen=5,
        ),
        readonly=False,
    )

    def read_value(self):
        arr2d = np.random.randint(0, 10, 5**2).reshape(5, 5)

        return arr2d.tolist()

    def write_arr3d(self, val):
        self.arr3d = val

    def read_arr3d(self):
        return self.arr3d


class Test_Struct_of_arrays(Readable):
    Status = Enum(Readable.Status)

    status = Parameter(datatype=StatusType(Status))

    value = Parameter(
        "struct of arrays containing primitive datatypes",
        datatype=StructOf(
            ints=ArrayOf(IntRange(), minlen=5, maxlen=5),
            strings=ArrayOf(StringType(), minlen=5, maxlen=5),
            floats=ArrayOf(FloatRange(), minlen=5, maxlen=5),
        ),
        readonly=True,
    )

    writable_strct_of_arr = Parameter(
        "writable struct of arrays containing primitive datatypes",
        datatype=StructOf(
            ints=ArrayOf(IntRange(), minlen=5, maxlen=5),
            strings=ArrayOf(StringType(), minlen=5, maxlen=5),
            floats=ArrayOf(FloatRange(), minlen=5, maxlen=5),
        ),
        readonly=False,
    )

    def read_value(self):
        strings = [
            "".join(random.choices(string.ascii_lowercase, k=5)) for _ in range(0, 5)
        ]
        ints = random.sample(range(0, 50), 5)
        floats = [random.random() for _ in range(0, 5)]

        return {"ints": ints, "strings": strings, "floats": floats}

    def read_writable_strct_of_arr(self):
        return self.writable_strct_of_arr

    def write_writable_strct_of_arr(self, val):
        self.writable_strct_of_arr = val


class OPHYD_test_primitive_arrays(Readable):

    value = Parameter("struct Value", datatype=FloatRange(), default=1.0, readonly=True)
    arr_int = Parameter(
        "array of ints",
        ArrayOf(IntRange(), 0, 10),
        readonly=True,
        default=[1, 2, 4, 5, 6, 7],
    )

    arr_float = Parameter(
        "array of floats",
        ArrayOf(FloatRange(), 0, 10),
        readonly=True,
        default=[1.2, 1.4],
    )

    arr_arr_float = Parameter(
        "array of floats",
        ArrayOf(ArrayOf(FloatRange(), 0, 10), 0, 10),
        readonly=True,
        default=[[2.3, 2.4], [1.2, 1.4]],
    )

    arr_bool = Parameter(
        "array of floats",
        ArrayOf(BoolType(), 0, 10),
        readonly=True,
        default=[True, False, True],
    )

    arr_String = Parameter(
        "array of Strings",
        ArrayOf(StringType(maxchars=20), 0, 10),
        readonly=True,
        default=["abdf", "sds", "ass"],
    )

    arr_String_nomax = Parameter(
        "array of Strings",
        ArrayOf(StringType(), 0, 10),
        readonly=True,
        default=["abdfefcsdcsdcsdcsdcsdcsdcsdc", "sds", "ass"],
    )


class OPYD_test_struct(Drivable):
    Status = Enum(Drivable.Status)  #: status codes

    status = Parameter(datatype=StatusType(Status))  # override Readable.status

    p_start = Parameter(
        "starting temperature for simulation",
        datatype=IntRange(0),
        default=5,
        export=False,
    )

    value = Parameter(
        "struct Value",
        datatype=StructOf(
            x=FloatRange(0, 100, unit="m"),
            y=FloatRange(0, 100, unit="m"),
            z=FloatRange(0, 100, unit="m"),
            color=StringType(),
        ),
    )

    target = Parameter(
        "struct target",
        datatype=StructOf(
            x=FloatRange(0, 100, unit="m"),
            y=FloatRange(0, 100, unit="m"),
            z=FloatRange(0, 100, unit="m"),
            color=StringType(),
        ),
    )

    nested_struct = Parameter(
        "nestedstruct dict containing other structs and tuples ",
        datatype=StructOf(
            number=FloatRange(0, 100, unit="s"),
            string=StringType(),
            tupl=TupleOf(FloatRange(0, 100), FloatRange(0, 100), FloatRange(0, 100)),
            pos_struct=StructOf(
                x=FloatRange(0, 100, unit="m"),
                y=FloatRange(0, 100, unit="m"),
                z=FloatRange(0, 100, unit="m"),
                col=StringType(),
                enum=EnumType(
                    "Test Enum", {"mode_zero": 0, "mode_one": 1, "mode_max": 2}
                ),
            ),
        ),
        readonly=False,
    )

    tuple_param = Parameter(
        "tuple parameter for testing",
        datatype=TupleOf(
            FloatRange(0, 100, unit="m"),
            FloatRange(0, 100, unit="m"),
            FloatRange(0, 100, unit="m"),
            StringType(),
        ),
        readonly=False,
    )

    tolerance = Parameter(
        "distance range for stability checking",
        datatype=FloatRange(0, 10),
        default=0.1,
        unit="m",
        readonly=False,
        group="stability",
    )

    looptime = Parameter(
        "timestep for simulation",
        datatype=FloatRange(0.01, 10),
        unit="s",
        default=1,
        readonly=False,
        export=False,
    )

    def read_value(self):
        return self.value

    def read_target(self):
        return self.target

    def read_status(self):
        # instead of asking a 'Hardware' take the value from the simulation
        return self.status

    def write_nested_struct(self, value):
        self.nested_struct = value

    def write_tuple_param(self, value):
        self.tuple_param = value

    def write_target(self, value):

        self.target = value
        # next read_status will see this status, until the loop updates it
        self.status = self.Status.BUSY, "new target set"
        return value

    @Command(
        StructOf(
            name=StringType(), id=IntRange(max=1000, min=0), sort=BoolType()
        ),
        result=IntRange(),
    )
    def test_cmd(self, name, id, sort):
        """testing with ophyd secop integration"""
        if name == "bad_name":
            raise ImpossibleError("bad name received")
        return random.randint(0, 1000)

    def initModule(self):
        super().initModule()
        self._stopflag = False
        self._thread = mkthread(self.thread)

    def thread(self):
        self.p_x = self.p_start
        self.p_y = self.p_start
        self.p_z = self.p_start

        self.t_x = self.p_start
        self.t_y = self.p_start
        self.t_z = self.p_start

        self.col = "blue"
        self.tcol = "blue"

        self.status = self.Status.IDLE, ""

        while not self._stopflag:
            try:
                self.__sim()
            except Exception as e:
                self.log.exception(e)
                self.status = self.Status.ERROR, str(e)

    def __heat(self, step, tolerance, target, pos):

        deviation = abs(pos - target)

        if deviation < tolerance:
            return pos, deviation

        if target > pos:
            return pos + step, deviation
        else:
            return pos - step, deviation

    def __sim(self):
        # complex thread handling:
        # a) simulation of cryo (heat flow, thermal masses,....)
        # b) optional PID temperature controller with windup control
        # c) generating status+updated value+ramp
        # this thread is not supposed to exit!

        # local state keeping:
        px = self.p_x
        py = self.p_y
        pz = self.p_z

        tx = self.t_x
        ty = self.t_y
        tz = self.t_z

        cols = ["blue", "red", "green", "yellow"]

        # keep history values for stability check

        timestamp = time.time()

        damper = 1

        while not self._stopflag:
            t = time.time()
            h = t - timestamp
            if h < self.looptime / damper:
                time.sleep(clamp(self.looptime / damper - h, 0.1, 60))
                continue
            # a)

            val = self.read_value()
            tar = self.read_target()

            px = val["x"]
            py = val["y"]
            pz = val["z"]

            tx = tar["x"]
            ty = tar["y"]
            tz = tar["z"]
            tcol = tar["color"]
            pcol = val["color"]

            new_px, dev_x = self.__heat(
                step=1, tolerance=self.tolerance, target=tx, pos=px
            )
            new_py, dev_y = self.__heat(
                step=1, tolerance=self.tolerance, target=ty, pos=py
            )
            new_pz, dev_z = self.__heat(
                step=1, tolerance=self.tolerance, target=tz, pos=pz
            )

            if tcol != pcol:
                new_col = cols[random.randint(0, 3)]
            else:
                new_col = pcol

            newpos = {"x": new_px, "y": new_py, "z": new_pz, "color": new_col}

            self.value = newpos
            # temperature is stable when all recorded values in the window
            # differ from setpoint by less than tolerance
            currenttime = time.time()

            # obtain min/max
            deviation = 0

            maxdev = abs(max([dev_x, dev_y, dev_z]))

            if maxdev > self.tolerance or pcol != tcol:
                self.status = self.Status.BUSY, "unstable"
            else:
                self.status = self.Status.IDLE, "at target"

            timestamp = t
            self.read_value()

    def shutdown(self):
        # should be called from server when the server is stopped
        self._stopflag = True
        if self._thread and self._thread.is_alive():
            self._thread.join()



# mypy: disable-error-code="attr-defined"
import numpy as np
from ophyd_async.core import SignalR, SignalRW

from secop_ophyd.AsyncFrappyClient import AsyncFrappyClient
from secop_ophyd.SECoPDevices import SECoPNodeDevice, SECoPReadableDevice


async def test_nested_connect(nested_struct_sim, nested_node: SECoPNodeDevice):
    assert isinstance(nested_node, SECoPNodeDevice)
    await nested_node.disconnect_async()


async def test_tuple_dev(nested_client: AsyncFrappyClient):
    ophy_struct = SECoPReadableDevice(
        secclient=nested_client, module_name="ophy_struct"
    )

    status_sig: SignalR = ophy_struct.status

    reading = await status_sig.read()

    reading_val = reading[status_sig.name]["value"]

    await status_sig.describe()

    stat0 = reading_val["f0"]
    stat1 = reading_val["f1"]

    assert stat0.item() == 300  # isinstance(stat0, int)

    assert isinstance(stat1.item(), str)

    await nested_client.disconnect(True)


async def test_struct_dev(nested_client: AsyncFrappyClient):
    ophy_struct = SECoPReadableDevice(
        secclient=nested_client, module_name="ophy_struct"
    )

    nested_struct_sig: SignalR = ophy_struct.nested_struct
    await nested_struct_sig.read()

    await nested_client.disconnect(True)


async def test_nested_dtype_str_signal_generation(
    nested_struct_sim, nested_node: SECoPNodeDevice
):
    struct_mod = nested_node.ophy_struct

    target: SignalRW = struct_mod.target

    reading = await target.read()

    descr_reading = await target.describe()

    descr = descr_reading.get(target.name)
    val = reading.get(target.name)["value"]

    assert isinstance(val, np.ndarray)
    assert descr["dtype"] == "array"
    assert descr["SECoP_dtype"] == "struct"
    await nested_node.disconnect_async()


async def test_nested_dtype_set_str_struct(
    nested_struct_sim, nested_node: SECoPNodeDevice
):
    struct_mod = nested_node.ophy_struct

    target: SignalRW = struct_mod.target

    reading = await target.read()

    val = reading.get(target.name)["value"]

    val["x"] = 20
    val["y"] = 30
    val["z"] = 40
    val["color"] = "yellow"

    stat = target.set(val)

    await stat

    reading = await target.read()

    val = reading.get(target.name)["value"]

    assert val["x"] == 20
    assert val["y"] == 30
    assert val["z"] == 40
    assert val["color"] == "yellow"
    assert isinstance(val, np.ndarray)

    await nested_node.disconnect_async()


async def test_nested_dtype_set_str_tuple(
    nested_struct_sim, nested_node: SECoPNodeDevice
):
    struct_mod = nested_node.ophy_struct

    tuple_param: SignalRW = struct_mod.tuple_param

    reading = await tuple_param.read()

    val = reading.get(tuple_param.name)["value"]

    assert val["f0"] == 5
    assert val["f1"] == 5
    assert val["f2"] == 5
    assert val["f3"] == "green"

    tuple_val = (50, 20, 30, "blue")

    stat = tuple_param.set(tuple_val)

    await stat

    reading = await tuple_param.read()

    val = reading.get(tuple_param.name)["value"]

    assert val["f0"] == 50
    assert val["f1"] == 20
    assert val["f2"] == 30
    assert val["f3"] == "blue"
    assert isinstance(val, np.ndarray)

    await nested_node.disconnect_async()


async def test_nested_struct_of_arrays(nested_struct_sim, nested_node: SECoPNodeDevice):
    str_of_arr_mod: SECoPReadableDevice = nested_node.struct_of_arrays

    reading = await str_of_arr_mod.read()

    val = reading[str_of_arr_mod.value.name]["value"]

    assert isinstance(val, np.ndarray)

    # Write testing
    rw_str_of_arr: SignalRW = str_of_arr_mod.writable_strct_of_arr

    rw_reading = await rw_str_of_arr.read()

    rw_val: np.ndarray = rw_reading[rw_str_of_arr.name]["value"]

    rw_old = rw_val.copy()

    rw_val["ints"] += 20
    rw_val["floats"] += 0.2

    await rw_str_of_arr.set(rw_val)

    rw_reading = await rw_str_of_arr.read()

    rw_val = rw_reading[rw_str_of_arr.name]["value"]

    assert np.equal(rw_val["ints"], rw_old["ints"] + 20).all()

    assert np.equal(rw_val["floats"], rw_old["floats"] + 0.2).all()

    await nested_node.disconnect_async()


# TODO Nested Arrays (2D) uniform and ragged



from __future__ import annotations

import copy
import time
import warnings
from abc import ABC, abstractmethod
from enum import Enum
from functools import reduce
from itertools import chain
from typing import Any, List, Union

import numpy as np
from bluesky.protocols import Reading
from event_model import DataKey
from frappy.client import CacheItem
from frappy.datatypes import (
    ArrayOf,
    BLOBType,
    BoolType,
    DataType,
    EnumType,
    FloatRange,
    IntRange,
    ScaledInteger,
    StringType,
    StructOf,
    TupleOf,
)
from ophyd_async.core._utils import StrictEnum

SCALAR_DATATYPES = (
    IntRange,
    BoolType,
    FloatRange,
    ScaledInteger,
    StringType,
    EnumType,
)


class Role(Enum):
    USER = 0
    ADVANCED = 1
    EXPERT = 2


class Access(Enum):
    NO_ACCESS = 0
    READ = 1
    WRITE = 2


def get_access_level(role: Role, accessmode: str) -> Access:
    assert len(accessmode) == 3

    access_str = accessmode[role.value]

    match access_str:
        case "r":
            return Access.READ
        case "w":
            return Access.WRITE
        case "-":
            return Access.NO_ACCESS
        case _:
            raise Exception(f"unknown accces level: {accessmode[role.value]}")


class NestedRaggedArray(Exception):
    """The Datatype contains nested ragged arrays"""


def deep_get(dictionary, keys, default=None) -> dict:
    def get_val(obj, key, default):
        if isinstance(obj, dict):
            return obj.get(key, default)
        if isinstance(obj, list):
            return obj[key]
        if isinstance(obj, tuple):
            return obj[key]
        return default

    return reduce(lambda d, key: get_val(d, key, default), keys, dictionary)


class Path:
    def __init__(self, parameter_name: str, module_name: str) -> None:
        self._accessible_name = parameter_name
        self._module_name = module_name
        self._last_named_param: int | None = None

        self._dev_path: List[Union[str, int]] = []

    # Path is extended
    def append(self, elem: Union[str, int]) -> Path:
        new_path = copy.deepcopy(self)

        if isinstance(elem, str):
            new_path._last_named_param = len(new_path._dev_path)

        new_path._dev_path.append(elem)

        return new_path

    def get_param_path(self):
        return {"module": self._module_name, "parameter": self._accessible_name}

    def get_path_tuple(self):
        return (self._module_name, self._accessible_name)

    # TODO typing of return value
    def get_memberinfo_path(self):
        # inserting K after every Nth number
        # using itertool.chain()

        # insert element
        k = "members"

        # insert after every other element
        n = 1

        # using itertool.chain()
        # inserting K after every Nth number
        return list(
            chain(
                *[
                    (
                        [k] + self._dev_path[i : i + n]  # type: ignore
                        if len(self._dev_path[i : i + n]) == n
                        else self._dev_path[i : i + n]
                    )
                    for i in range(0, len(self._dev_path), n)
                ]
            )
        )

    def get_signal_name(self):
        # top level: signal name == Parameter name

        if self._dev_path == []:
            return self._accessible_name

        sig_postfix = self._dev_path[self._last_named_param :]

        if self._last_named_param is None:
            sig_postfix = [self._accessible_name] + sig_postfix  # type: ignore

        delim = "-"
        return delim.join(map(str, sig_postfix))

    def get_param_desc_path(self):
        return [self._module_name, "parameters", self._accessible_name]

    def get_cmd_desc_path(self):
        return [self._module_name, "commands", self._accessible_name]

    def get_leaf(self):
        if self._dev_path == []:
            return None

        return self._dev_path[-1]

    def insert_val(self, dic: dict, new_val):
        if self._dev_path == []:
            return dic

        d = dic
        for key in self._dev_path[:-1]:
            if key in d:
                d = d[key]
            else:
                # wrong path
                raise Exception(
                    "path is incorrect " + str(key) + " is not in dict: " + str(dic)
                )
        # insert new value
        if self._dev_path[-1] in d:
            d[self._dev_path[-1]] = new_val
        else:
            # wrong path
            raise Exception(
                "path is incorrect " + str(key) + " is not in dict: " + str(dic)
            )
        return dic


class DtypeNP(ABC):
    secop_dtype: DataType
    name: str | None
    array_element: bool = False
    max_depth: int = 0

    @abstractmethod
    def make_numpy_dtype(self) -> tuple:
        """Create Numpy Compatible structured Datatype"""

    @abstractmethod
    def make_concrete_numpy_dtype(self, value) -> tuple:
        """Create Numpy Compatible structured Datatype from a concrete data value"""

    @abstractmethod
    def make_numpy_compatible_list(self, value) -> Any:
        """make a make a list that is ready for numpy array import"""

    @abstractmethod
    def make_secop_compatible_object(self, value) -> Any:
        """make a make an SECoP Compatible Object"""


def dt_factory(
    secop_dt: DataType, name: str = "", array_element: bool = False
) -> DtypeNP:
    dt_class = secop_dt.__class__

    dt_converters = {
        StructOf: StructNP,
        TupleOf: TupleNP,
        ArrayOf: ArrayNP,
        IntRange: IntNP,
        FloatRange: FloatNP,
        ScaledInteger: ScaledIntNP,
        BLOBType: BLOBNP,
        BoolType: BoolNP,
        EnumType: EnumNP,
        StringType: StringNP,
    }

    return dt_converters[dt_class](secop_dt, name, array_element)  # type: ignore


STR_LEN_DEFAULT = 100


class BLOBNP(DtypeNP):
    def __init__(
        self, blob_dt: BLOBType, name: str = "", array_element: bool = False
    ) -> None:
        self.name: str = name
        self.secop_dtype: BLOBType = blob_dt
        self.array_element = array_element

    def make_numpy_dtype(self) -> tuple:
        return (self.name, "U" + str(self.secop_dtype.maxbytes))

    def make_concrete_numpy_dtype(self, value) -> tuple:
        return (self.name, "U" + str(self.secop_dtype.maxbytes))

    def make_numpy_compatible_list(self, value: str):
        return value

    def make_secop_compatible_object(self, value) -> Any:
        return value


class BoolNP(DtypeNP):
    def __init__(
        self, bool_dt: BoolType, name: str = "", array_element: bool = False
    ) -> None:
        self.name: str = name
        self.secop_dtype: BoolType = bool_dt
        self.array_element = array_element

    def make_numpy_dtype(self) -> tuple:
        return (self.name, "<b1")

    def make_concrete_numpy_dtype(self, value) -> tuple:
        return self.make_numpy_dtype()

    def make_numpy_compatible_list(self, value: bool):
        return value

    def make_secop_compatible_object(self, value) -> Any:
        return value


class EnumNP(DtypeNP):
    def __init__(
        self, enum_dt: EnumType, name: str = "", array_element: bool = False
    ) -> None:
        self.name: str = name
        self.secop_dtype: EnumType = enum_dt
        self.array_element = array_element

    def make_numpy_dtype(self) -> tuple:
        return (self.name, "<i8")

    def make_concrete_numpy_dtype(self, value) -> tuple:
        return self.make_numpy_dtype()

    def make_numpy_compatible_list(self, value: int):
        return value

    def make_secop_compatible_object(self, value) -> Any:
        return value


class FloatNP(DtypeNP):
    def __init__(
        self, float_dt: FloatRange, name: str = "", array_element: bool = False
    ) -> None:
        self.name: str = name
        self.secop_dtype: FloatRange = float_dt
        self.array_element = array_element

    def make_numpy_dtype(self) -> tuple:
        return (self.name, "<f8")

    def make_concrete_numpy_dtype(self, value) -> tuple:
        return self.make_numpy_dtype()

    def make_numpy_compatible_list(self, value: float):
        return value

    def make_secop_compatible_object(self, value) -> Any:
        return value


class IntNP(DtypeNP):
    def __init__(
        self, int_dt: IntRange, name: str = "", array_element: bool = False
    ) -> None:
        self.name: str = name
        self.secop_dtype: IntRange = int_dt
        self.array_element = array_element

    def make_numpy_dtype(self) -> tuple:
        return (self.name, "<i8")

    def make_concrete_numpy_dtype(self, value) -> tuple:
        return self.make_numpy_dtype()

    def make_numpy_compatible_list(self, value: int):
        return value

    def make_secop_compatible_object(self, value) -> Any:
        return value


class ScaledIntNP(DtypeNP):
    def __init__(
        self, scaled_int_dt: ScaledInteger, name: str = "", array_element: bool = False
    ) -> None:
        self.name: str = name
        self.secop_dtype: ScaledInteger = scaled_int_dt
        self.array_element = array_element

    def make_numpy_dtype(self) -> tuple:
        return (self.name, "<i8")

    def make_concrete_numpy_dtype(self, value) -> tuple:
        return self.make_numpy_dtype()

    def make_numpy_compatible_list(self, value: int):
        return value

    def make_secop_compatible_object(self, value) -> Any:
        return value


class StringNP(DtypeNP):
    def __init__(
        self, string_dt: StringType, name: str = "", array_element: bool = False
    ) -> None:
        self.name: str = name
        self.secop_dtype: StringType = string_dt
        self.array_element = array_element

        if string_dt.maxchars == 1 << 64:
            # warnings.warn(
            #    "maxchars was not set, default max char lenght is set to: "
            #    + str(STR_LEN_DEFAULT)
            # )
            self.strlen = STR_LEN_DEFAULT

        else:
            self.strlen = string_dt.maxchars

    def make_numpy_dtype(self) -> tuple:
        return (self.name, "<U" + str(self.strlen))

    def make_concrete_numpy_dtype(self, value) -> tuple:
        return (self.name, "<U" + str(self.strlen))

    def make_numpy_compatible_list(self, value: str):
        return value

    def make_secop_compatible_object(self, value) -> Any:
        return value


class StructNP(DtypeNP):
    def __init__(
        self, struct_dt: StructOf, name: str = "", array_element: bool = False
    ) -> None:
        self.name: str = name
        self.secop_dtype: StructOf = struct_dt
        self.array_element = array_element
        self.members: dict[str, DtypeNP] = {
            name: dt_factory(member, name, self.array_element)
            for (name, member) in struct_dt.members.items()
        }

        max_depth = [member.max_depth for member in self.members.values()]
        self.max_depth = 1 + max(max_depth)

    def make_numpy_dtype(self) -> tuple:
        dt_list = []
        for member in self.members.values():
            dt_list.append(member.make_numpy_dtype())

        return (self.name, dt_list)

    def make_concrete_numpy_dtype(self, value) -> tuple:
        dt_list = []
        for key, member in self.members.items():

            member_val = value[key]
            dt_list.append(member.make_concrete_numpy_dtype(member_val))

        return (self.name, dt_list)

    def make_numpy_compatible_list(self, value: dict):
        return tuple(
            [
                self.members[name].make_numpy_compatible_list(value[name])
                for name in self.members.keys()
            ]
        )

    def make_secop_compatible_object(self, value) -> Any:
        return {
            member: np_dtype.make_secop_compatible_object(value=value[member])
            for (member, np_dtype) in self.members.items()
        }


class TupleNP(DtypeNP):
    def __init__(
        self, tuple_dt: TupleOf, name: str = "", array_element: bool = False
    ) -> None:
        self.name: str = name
        self.secop_dtype = tuple_dt
        self.array_element = array_element
        self.members: list[DtypeNP] = [
            dt_factory(
                secop_dt=member, name="f" + str(idx), array_element=self.array_element
            )
            for idx, member in enumerate(tuple_dt.members)
        ]

        max_depth = [member.max_depth for member in self.members]
        self.max_depth = 1 + max(max_depth)

    def make_numpy_dtype(self) -> tuple:
        dt_list = []
        for member in self.members:
            dt_list.append(member.make_numpy_dtype())

        return (self.name, dt_list)

    def make_concrete_numpy_dtype(self, value) -> tuple:
        dt_list = []
        for index, member in enumerate(self.members):
            member_val = value[index]
            dt_list.append(member.make_concrete_numpy_dtype(member_val))

        return (self.name, dt_list)

    def make_numpy_compatible_list(self, value: tuple):
        return tuple(
            [
                self.members[idx].make_numpy_compatible_list(value[idx])
                for idx, member in enumerate(self.members)
            ]
        )

    def make_secop_compatible_object(self, value) -> Any:
        return tuple(
            [
                np_dtype.make_secop_compatible_object(value=value["f" + str(idx)])
                for idx, np_dtype in enumerate(self.members)
            ]
        )


class ArrayNP(DtypeNP):
    def __init__(
        self, array_dt: ArrayOf, name: str = "", array_element: bool = False
    ) -> None:
        self.name: str = name
        self.secop_dtype = array_dt
        self.maxlen = array_dt.maxlen
        self.minlen = array_dt.minlen
        self.array_element = array_element

        self.ragged: bool = self.minlen != self.maxlen

        self.shape = [self.maxlen]

        self.members: DtypeNP = dt_factory(array_dt.members, array_element=True)
        self.root_type: DtypeNP

        self.max_depth = self.members.max_depth

        if isinstance(self.members, ArrayNP):
            self.shape.extend(self.members.shape)
            self.root_type = self.members.root_type
            self.members.shape = []
            if self.members.ragged:
                pass
                # raise NestedRaggedArray(
                #    "ragged arrays with more than a single dimension are not supported"
                # )int

        else:
            self.root_type = self.members

        if self.array_element and self.ragged:
            warnings.warn(
                "ragged arrays inside of arrays of copmposite datatypes (struct/tuple)"
                "are not supported"
            )

    def get_root_np_str(self) -> str:
        dtype_list = self.root_type.make_numpy_dtype()
        return dtype_list[1]

    def make_numpy_dtype(self) -> tuple:
        if self.shape == []:
            return self.members.make_numpy_dtype()
        else:
            return (self.name, list(self.members.make_numpy_dtype()).pop(), self.shape)

    def make_concrete_numpy_dtype(self, value) -> tuple:

        if self.ragged is False:
            return (
                self.name,
                list(self.members.make_concrete_numpy_dtype(value)).pop(),
                self.shape,
            )
        else:
            if value == []:
                member_np = self.members.make_concrete_numpy_dtype(None)
                val_shape = [0]
            elif value is None:
                member_np = self.members.make_concrete_numpy_dtype(None)
                val_shape = []
            else:
                member_np = self.members.make_concrete_numpy_dtype(value[0])
                val_shape = [len(value)]

            if isinstance(self.members, ArrayNP):
                val_shape = val_shape + member_np[2]

            return (self.name, member_np[1], val_shape)

    def make_numpy_compatible_list(self, value: list):
        return [self.members.make_numpy_compatible_list(elem) for elem in value]

    def make_secop_compatible_object(self, value: np.ndarray) -> Any:
        return [
            self.members.make_secop_compatible_object(value[idx])
            for idx in range(0, self.maxlen)
        ]


class SECoPdtype:

    def __init__(self, datatype: DataType) -> None:
        self.raw_dtype: DataType = datatype

        # Describe Fields ------------------------

        self.dtype: str

        # The array-protocol typestring of the data-type object.
        self.dtype_str: str

        # String representation of the numpy structured array dtype
        self.dtype_descr: list

        # string representation of the original secop datatype
        self.secop_dtype_str = datatype.export_datatype()

        # Shape of Data
        self.shape = []

        self.np_datatype: Any

        # Describe Fields ------------------------

        self.numpy_dtype: np.dtype

        self.dtype_tree: DtypeNP

        self._is_composite: bool = False
        self._is_array: bool = False

        self.dtype_tree = dt_factory(datatype)

        self.max_depth: int = self.dtype_tree.max_depth

        if isinstance(self.dtype_tree, ArrayNP):
            self.shape = self.dtype_tree.shape
            self._is_array = True
            self._is_composite = (
                True
                if isinstance(self.dtype_tree.root_type, (StructNP, TupleNP))
                else False
            )

        if isinstance(self.dtype_tree, (TupleNP, StructNP)):
            self._is_composite = True

        # Composite Datatypes & Arrays of COmposite Datatypes
        if self._is_composite:
            dt = self.dtype_tree.make_numpy_dtype()

            # Top level elements are not named and shape is
            # already covered by the shape var
            dt = dt[1]

            self.numpy_dtype = np.dtype(dt)

            # all composite Dtypes are transported as numpy arrays
            self.dtype = "array"
            self.dtype_str = self.numpy_dtype.str
            self.dtype_descr = self.numpy_dtype.descr
            self.np_datatype = np.ndarray

        # Scalar atomic Datatypes and arrays of atomic dataypes
        else:
            if self._is_array:
                # root secop datatype that is contained in the array
                self.dtype = "array"
                self.np_datatype = np.ndarray

            # Primitive datatypes
            else:
                self.np_datatype = SECOP2DTYPE[datatype.__class__][0]
                self.dtype = SECOP2DTYPE[datatype.__class__][1]

    def get_datakey(self) -> DataKey:
        describe_dict: dict = {}
        # Composite Datatypes & Arrays of COmposite Datatypes
        if self._is_composite:
            describe_dict["dtype_str"] = self.dtype_str
            # describe_dict["dtype_numpy"] = self.dtype_descr
            describe_dict["dtype_descr"] = self.dtype_descr

        if isinstance(self.dtype_tree, ArrayNP):
            describe_dict["dtype_numpy"] = self.dtype_tree.get_root_np_str()

        describe_dict["dtype"] = self.dtype
        describe_dict["shape"] = self.shape
        describe_dict["SECOP_datainfo"] = self.secop_dtype_str

        return describe_dict

    def _secop2numpy_array(self, value) -> np.ndarray:
        np_list = self.dtype_tree.make_numpy_compatible_list(value)

        return np.array(np_list, dtype=self.numpy_dtype)

    def secop2val(self, reading_val) -> Any:
        if self._is_composite:
            return self._secop2numpy_array(reading_val)

        else:
            return reading_val

    def val2secop(self, input_val) -> Any:
        # TODO check input_Val for conformity with datatype
        # TODO check if it is already in SECoP Format

        if self._is_composite and isinstance(input_val, np.ndarray):
            return self.dtype_tree.make_secop_compatible_object(input_val)
        else:
            return self.raw_dtype.validate(input_val)

    def update_dtype(self, input_val):
        if self._is_composite:
            # Composite Datatypes & Arrays of Composite Datatypes

            dt = self.dtype_tree.make_concrete_numpy_dtype(input_val)

            # Top level elements are not named and shape is
            # already covered by the shape var
            inner_dt = dt[1]

            if isinstance(self.raw_dtype, ArrayOf):
                self.shape = dt[2]

            self.numpy_dtype = np.dtype(inner_dt)

            self.dtype_str = self.numpy_dtype.str
            self.dtype_descr = self.numpy_dtype.descr

            return

        if isinstance(self.raw_dtype, ArrayOf):
            dt = self.dtype_tree.make_concrete_numpy_dtype(input_val)

            self.shape = dt[2]


class SECoPReading:
    def __init__(
        self,
        secop_dt: SECoPdtype,
        entry: CacheItem | None = None,
    ) -> None:
        self.secop_dt: SECoPdtype = secop_dt

        if entry is None:
            self.timestamp: float = time.time()
            self.value = None
            self.readerror = None

            return

        if entry.readerror is not None:
            raise entry.readerror

        exported_val = secop_dt.raw_dtype.export_value(entry.value)

        self.secop_dt.update_dtype(exported_val)

        self.value = secop_dt.secop2val(exported_val)

        self.secop_val = exported_val

        self.timestamp = entry.timestamp

    def get_reading(self) -> Reading:
        return {"value": self.value, "timestamp": self.timestamp}

    def get_value(self):
        return self.value

    def get_secop_value(self):
        return self.secop_val

    def set_reading(self, value) -> None:
        self.value = value
        self.secop_val = self.secop_dt.val2secop(value)
        self.timestamp = time.time()


SECOP2DTYPE = {
    FloatRange: (float, "number"),
    IntRange: (int, "integer"),
    ScaledInteger: (int, "integer"),
    BoolType: (bool, "boolean"),
    EnumType: (StrictEnum, "string"),
    StringType: (str, "string"),
    BLOBType: (str, "string"),
}
